/*
 * $Id: e1372d414695b6f63b0ec40f5ef547f6eb1363c0 $
 *
 * This file is part of ntldap.
 *
 * Copyright (c) 2006, 2007, 2008 Niklas Olmes <niklas@noxa.de>
 *                                http://www.nevm.de/ntldap/
 *                                NEVM-RIPE
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <signal.h>
#include <time.h>
#include <limits.h>
#include <termcap.h>
#ifdef __linux__
#include <values.h>
#endif

#include "ncurses.h"
#include "menu.h"
#include "form.h"
#include "cdk/cdk.h"

#include "ldap.h"

#include "ntg.h"
#include "ntg_keys.h"
#include "ntg_cdkmenu.h"
#include "ntl.h"
#include "ntlc.h"
#include "ntls.h"
#include "ntu.h"
#include "nt_alloc.h"

#define NTG_MAXFIELDS			196
#define NTG_GENFIELDS_BOTTOM_SPACE	3

static FIELD *f_label(const char *, int, int, int);
static FIELD *f_field(const char *, const char *, int, int, int, int, int, int, const char *, const char **);
static chtype ntg_genfields_mapcolor(int, int);
static int ntg_input_keybind(EObjectType, void *, void *, chtype);
static int ntg_select_keybind(EObjectType, void *, void *, chtype);
static int ntg_select_multi_keybind(EObjectType, void *, void *, chtype);

static void ntg_sigwinch(int);

extern const char *ntl_error_str;
extern const char *ntls_error_str;
extern int ntl_error;

static const char rcsid[] = "$Id: e1372d414695b6f63b0ec40f5ef547f6eb1363c0 $";

static const char myname[] = "nldap";
static const char release[] = NTLDAP_VERSION;

static const char *lda_search[] = { "*", "hasSubordinates", NULL };
static const char *lda_showop[] = { "*", "hasSubordinates", "+", NULL };
static const char *boollist[] = {"FALSE", "TRUE", NULL };

static const char *quest_ays[] = { "Are You Sure? (y/n)", NULL };
static const char *fval_msg[] = { "Validation Failed.", NULL };
static const char *cr_msg[] = { "Creating...", NULL };
static const char *dl_msg[] = { "Deleting Entry...", NULL };
static const char *wr_msg[] = { "Writing Entry...", NULL };
static const char *mod_msg[] = { "Modifying Entry...", NULL };
static const char *cp_msg[] = { "Copying Entry...", NULL };
static const char *mv_msg[] = { "Moving Entry...", NULL };
static const char *retr_msg[] = { "Getting Entry...", NULL };
static const char *retr_p_msg[] = { "Getting Entries...", NULL };
static const char *nf_msg[] = { "Nothing Found.", NULL };
static const char *ie_msg[] = { "Internal Error.", NULL };
static const char *ntplf_msg[] = { "No Suitable Template For Displaying This Object Found.", NULL };
static const char *ntplc_msg[] = { "No Suitable Template For Creating This Object Found.", NULL };
static const char *nfu_msg[] = { "Cannot Go Further.", NULL };
static const char *nval_msg[] = { "Need Values.", NULL };
static const char *c_msg[] = { "Connecting To LDAP...", NULL };
static const char *rschema_msg[] = { "Reading Schema From LDAP...", NULL };
static const char *nschema_msg[] = { "Need Schema From LDAP For This To Work.", NULL };
static const char *editor_msg[] = { "Launching Editor...", NULL };

static const char inp_pass[] = "Input Password: ";
static const char inp_newdn[] = "Enter DN Of New Entry: ";
static const char inp_newloc[] = "Enter DN Of New Location: ";
static const char inp_targetbase[] = "Enter Target Base DN: ";

static char cdk_select_bgcolor[] = "</33>";
static char cdk_input_bgcolor[] = "</29>";
static char cdk_display_bgcolor[] = "</10>";
static char cdk_tip_bgcolor[] = "</60>";
static char cdk_info_bgcolor[] = "</60>";
static char cdk_yesno_bgcolor[] = "</58>";
static char cdk_input_str[] = "[</B>Input<!B>]";

static NTG_CTX *_ctx_ = NULL;

void
ntg_fail(NTG_CTX *ctx, const char *s, int e)
{
	ntg_ctx_destroy(ctx);
	ntg_cleanup();
	errx(e, "ntg_fail: (%d) %s", e, s);
}

void
ntg_cleanup(void)
{
	endCDK();
	(void)clear();
	(void)endwin();
}

NTG_CTX *
ntg_ctx_init(void)
{
	NTG_CTX *ctx;

	ctx = nt_calloc(1, sizeof(NTG_CTX));
	ctx->rootwin = NULL;
	ctx->topwin = NULL;
	ctx->mainwin = NULL;
	ctx->bottomwin = NULL;
	ctx->cdkscreen = NULL;
	ctx->cdkrootscreen = NULL;
	ctx->cdklabel = NULL;
	ctx->cdktip = NULL;
	ctx->cdkmenus = NULL;
	ctx->ld = NULL;
	ctx->cfg = NULL;
	ctx->schema = NULL;
	ctx->keys = NULL;
	ctx->tmp = NULL;
	ctx->tmpflags = 0;
	ctx->tmpreturn = 0;
	ctx->keys = NULL;
	ctx->sortleft = 1;
	ctx->mainwin_maxx = ctx->mainwin_maxy = 0;
	ctx->stdscr_maxx = ctx->stdscr_maxy = 0;
	*ctx->title = '\0';
	*ctx->subtitle = '\0';
	ctx->keyhelp = NULL;

	return ctx;
}

void
ntg_ctx_destroy(NTG_CTX *ctx)
{
	int i;

	if (ctx == NULL)
		return;

	if (ctx->cdkmenus != NULL) {
		for (i = 0; ctx->cdkmenus[i] != NULL; i++)
			ntg_cdkmenu_free(ctx->cdkmenus[i]);
		free(ctx->cdkmenus);
	}
	if (ctx->cdklabel != NULL)
		destroyCDKLabel(ctx->cdklabel);
	if (ctx->cdktip != NULL)
		destroyCDKLabel(ctx->cdktip);
	if (ctx->cdkscreen != NULL)
		destroyCDKScreen(ctx->cdkscreen);
	if (ctx->cdkrootscreen != NULL)
		destroyCDKScreen(ctx->cdkrootscreen);
	if (ctx->topwin != NULL)
		delwin(ctx->topwin);
	if (ctx->mainwin != NULL)
		delwin(ctx->mainwin);
	if (ctx->bottomwin != NULL)
		delwin(ctx->bottomwin);
	if (ctx->rootwin != NULL)
		delwin(ctx->rootwin);
	if (ctx->ld != NULL)
		(void)ldap_unbind(ctx->ld);
	if (ctx->schema != NULL)
		ntls_schema_free(ctx->schema);
	if (ctx->cfg != NULL)
		ntlc_cfg_free(ctx->cfg);
	if (ctx->tmp != NULL)
		free(ctx->tmp);
	if (ctx->keys != NULL)
		ntg_keys_cfg_free(ctx->keys);
	free(ctx);
	ctx = NULL;
	_ctx_ = NULL;
}

int
ntg_init_curses(NTG_CTX *ctx)
{

	ctx->rootwin = initscr();

	ctx->stdscr_maxx = getmaxx(stdscr);
	ctx->stdscr_maxy = getmaxy(stdscr);
	if (ctx->stdscr_maxy < NTG_MIN_HEIGHT || ctx->stdscr_maxx < NTG_MIN_WIDTH) {
		endwin();
		errx(1, "Need a minimum of %dx%d chars on terminal to work.", NTG_MIN_HEIGHT, NTG_MIN_WIDTH);
	}

	if (!has_colors())
		warnx("colors unavailable on this terminal");

	if ((ctx->topwin = newwin(2, ctx->stdscr_maxx, 0, 0)) == NULL)
		ntg_fail(ctx, "unable to create top window", NTG_E_INIT);

	if ((ctx->mainwin = newwin(ctx->stdscr_maxy - 2 - (NTG_KEYHELP_LINES + 1), ctx->stdscr_maxx - 4, 2, 2)) == NULL)
		ntg_fail(ctx, "unable to create main window", NTG_E_INIT);
	ctx->mainwin_maxx = getmaxx(ctx->mainwin);
	ctx->mainwin_maxy = getmaxy(ctx->mainwin);

	if ((ctx->bottomwin = newwin(NTG_KEYHELP_LINES + 1, ctx->stdscr_maxx, ctx->stdscr_maxy - NTG_KEYHELP_LINES - 1, 0)) == NULL)
		ntg_fail(ctx, "unable to create bottom window", NTG_E_INIT);

	ctx->cdkscreen = initCDKScreen(ctx->mainwin);
	ctx->cdkrootscreen = initCDKScreen(ctx->rootwin);

	_ctx_ = ctx;

	(void)use_default_colors();
	(void)start_color();
	initCDKColor();
	(void)init_pair(C_NORMAL_NUM, COLOR_WHITE, C_DEFAULT_BG);
	(void)init_pair(C_TEXT_NUM, COLOR_CYAN, C_DEFAULT_BG);
	(void)init_pair(C_STRONG_NUM, COLOR_YELLOW, C_DEFAULT_BG);
	(void)init_pair(C_EM_NUM, COLOR_GREEN, C_DEFAULT_BG);
	(void)init_pair(C_SUCCESS_NUM, COLOR_GREEN, C_DEFAULT_BG);
	(void)init_pair(C_ERROR_NUM, COLOR_RED, C_DEFAULT_BG);
	(void)init_pair(C_INPUT_NUM, COLOR_WHITE, C_DEFAULT_BG);
	(void)init_pair(C_INFO_NUM, COLOR_YELLOW, C_DEFAULT_BG);
	(void)init_pair(C_MSG_NUM, COLOR_WHITE, C_DEFAULT_BG);
	(void)init_pair(C_BLUE_NUM, COLOR_BLUE, COLOR_WHITE);
	(void)wbkgd(ctx->mainwin, C_NORMAL);

	ntg_cdkmenu_init(ctx);

	ntu_seed();

	(void)atexit(ntg_cleanup);

	(void)signal(SIGWINCH, ntg_sigwinch);

	return 0;
}

static void
ntg_sigwinch(int dummy)
{
	char buf[1024];
	char li[] = "li";
	char co[] = "co";

	(void)dummy; /* unused */

	if (_ctx_ == NULL)
		return;

	/*
	 * NOTE/XXX:
	 * Everything we do here is unsafe.
	 */

	if (tgetent(buf, getenv("TERM")) < 0)
		return;

	_ctx_->stdscr_maxy = tgetnum(li);
	_ctx_->stdscr_maxx = tgetnum(co);
	(void)resizeterm(_ctx_->stdscr_maxy, _ctx_->stdscr_maxx);
	(void)wresize(stdscr, _ctx_->stdscr_maxy, _ctx_->stdscr_maxx);

	_ctx_->mainwin_maxy = _ctx_->stdscr_maxy - 2 - (NTG_KEYHELP_LINES + 1);
	_ctx_->mainwin_maxx = _ctx_->stdscr_maxx - 4;
	(void)wresize(_ctx_->mainwin, _ctx_->mainwin_maxy, _ctx_->mainwin_maxx);
	(void)wresize(_ctx_->topwin, 2, _ctx_->stdscr_maxx);
	(void)wresize(_ctx_->bottomwin, NTG_KEYHELP_LINES + 1, _ctx_->stdscr_maxx);
	(void)mvwin(_ctx_->bottomwin, _ctx_->stdscr_maxy - NTG_KEYHELP_LINES - 1, 0);

	(void)wclear(stdscr);
	(void)wclear(_ctx_->topwin);
	(void)wclear(_ctx_->mainwin);
	(void)wclear(_ctx_->bottomwin);

	ntg_title(_ctx_, *_ctx_->title == '\0' ? NULL : _ctx_->title);
	ntg_subtitle(_ctx_, *_ctx_->subtitle == '\0' ? NULL : _ctx_->subtitle);
	ntg_keyhelp(_ctx_, _ctx_->keyhelp);

	(void)mvwprintw(_ctx_->mainwin, _ctx_->mainwin_maxy / 2, _ctx_->mainwin_maxx / 2 - 12, "Window Size Has Changed:");
	(void)mvwprintw(_ctx_->mainwin, _ctx_->mainwin_maxy / 2 + 1, _ctx_->mainwin_maxx / 2 - 27, "You will have to reload or otherwise force a refresh.");
	ntg_refresh(_ctx_);
}

int
ntg_init_ldap(NTG_CTX *ctx)
{

	ntg_info_up(ctx, c_msg);
	if ((ctx->ld = ntl_connect(ctx->cfg)) == NULL) {
		ntg_error_va(ctx, "LDAP Connect Failed:\n%s", ntl_error_str);
		return 1;
	}
	ntg_info_down(ctx);

	return 0;
}

int
ntg_init_schema(NTG_CTX *ctx)
{

	ntg_info_up(ctx, rschema_msg);
	if ((ctx->schema = ntls_read_schema(ctx->ld)) == NULL) {
		ntg_error_va(ctx, "Failure Reading LDAP Schema:\n%s", ntls_error_str);
		return 1;
	}
	ntg_info_down(ctx);

	return 0;
}

static int
ntg_select_keybind(EObjectType cdkType, void *object, void *clientData, chtype key)
{
	CDKSCROLL *s;

	/* unused */
	(void)clientData;
	(void)cdkType;

	s = object;
	if (key == KEY_HOME)
		(void)injectCDKScroll(s, (chtype)'1');
	else if (key == KEY_END)
		(void)injectCDKScroll(s, (chtype)'G');
	else
		return 1;

	return 0;
}

int
ntg_select(NTG_CTX *ctx, const char *title, const char **p_choices)
{
	CDKSCROLL *s;
	char **choices;
	int selection;
	int h, v;
	int i, count;

	count = 0;
	while (p_choices[count] != NULL)
		count++;

	choices = nt_calloc((size_t)count + 1, sizeof(char *));
	for (i = 0; i < count; i++)
		choices[i] = nt_strdup(p_choices[i]);
	choices[count] = NULL;

	h = ctx->mainwin_maxx * 5/10;
	v = ctx->mainwin_maxy * 5/10;
	if (h < 30)
		h = 30;
	if (v < 10)
		v = 10;

	if ((s = newCDKScroll(ctx->cdkscreen, CENTER, CENTER, RIGHT, v, h, (char *)title, choices, count, NUMBERS, A_REVERSE, TRUE, FALSE)) == NULL)
		selection = -1;
	else {
		setCDKScrollBackgroundColor(s, cdk_select_bgcolor);
		if (CDK_VERSION_MAJOR == "4")
			setCDKScrollPreProcess(s, ntg_select_keybind, NULL);

		selection = activateCDKScroll(s, NULL);
		if (selection < 0)
			selection = -2;
		else if (selection > count)
			selection = -3;
	}

	if (s != NULL) {
		eraseCDKScroll(s);
		destroyCDKScroll(s);
	}
	refreshCDKScreen(ctx->cdkscreen);

	while (count > 0)
		free(choices[--count]);
	free(choices);

	return selection;
}

static int
ntg_select_multi_keybind(EObjectType cdkType, void *object, void *clientData, chtype key)
{
	CDKSELECTION *sel;

	/* unused */
	(void)clientData;
	(void)cdkType;

	sel = object;
	if (key == KEY_HOME)
		(void)injectCDKSelection(sel, (chtype)'1');
	else if (key == KEY_END)
		(void)injectCDKSelection(sel, (chtype)'G');
	else
		return 1;

	return 0;
}

int *
ntg_select_multi(NTG_CTX *ctx, const char *title, const char **p_choices)
{
	CDKSELECTION *sel;
	const char *c_str[] = { "   ", "(*) ", NULL };
	char **choices;
	int *selection;
	int h, v;
	int i, count;

	count = 0;
	while (p_choices[count] != NULL)
		count++;

	choices = nt_calloc((size_t)count + 1, sizeof(char *));
	for (i = 0; i < count; i++)
		choices[i] = nt_strdup(p_choices[i]);
	choices[count] = NULL;

	h = ctx->mainwin_maxx * 5/10;
	v = ctx->mainwin_maxy * 5/10;
	if (h < 30)
		h = 30;
	if (v < 12)
		v = 12;

	if ((sel = newCDKSelection(ctx->cdkscreen, CENTER, CENTER, RIGHT, v, h, (char *)title, choices, count, (char **)c_str, 2, A_REVERSE, TRUE, FALSE)) == NULL)
		selection = NULL;
	else {
		setCDKSelectionBackgroundColor(sel, cdk_select_bgcolor);
		if (CDK_VERSION_MAJOR == "4")
			setCDKSelectionPreProcess(sel, ntg_select_multi_keybind, NULL);

		if (activateCDKSelection(sel, NULL) != 1)
			selection = NULL;
		else {
			selection = nt_calloc((size_t)count, sizeof(int));
			for (i = 0; i < count; i++)
				selection[i] = sel->selections[i];
		}
	}

	if (sel != NULL) {
		eraseCDKSelection(sel);
		destroyCDKSelection(sel);
	}
	refreshCDKScreen(ctx->cdkscreen);

	while (count > 0)
		free(choices[--count]);
	free(choices);

	return selection;
}

int
ntg_newobject_tpl(NTG_CTX *ctx, const char *basedn)
{
	char **tplnames;
	char *newbasedn;
	int choice;
	int r;
	int i, count;

	count = 0;
	while (ctx->cfg->tplc[count] != NULL)
		count++;

	tplnames = nt_calloc((size_t)count + 1, sizeof(char *));
	for (i = 0; i < count; i++)
		tplnames[i] = ctx->cfg->tplc[i]->name;
	tplnames[count] = NULL;

	choice = ntg_select(ctx, "Choose A Template: ", (const char **)tplnames);
	free(tplnames);

	if (choice < 0)
		return -1;

	if (*(newbasedn = ntg_input(ctx, "Base DN (Return For Default): ", basedn)) == '\0') {
		free(newbasedn);
		return -2;
	}

	r = ntg_create_obj_from_tpl(ctx, newbasedn, ctx->cfg->tplc[choice]);
	free(newbasedn);
	return r;
}

int
ntg_newobject_oc(NTG_CTX *ctx, const char *basedn)
{
	char **ocs_abs, **ocs_aux, **ocs_str;
	char **r_ocs, **i_ocs;
	char **req_ats;
	size_t c_abs, c_aux, c_str;
	int *choices;
	int r_count;
	int flag;
	int e;
	unsigned int i;

	/* unused */
	(void)ctx;
	(void)basedn;

	/*
	 * XXX/GCC: GCC3.4 Bug under Gentoo Hardened with -O2
	 * warning: 'req_ats' might be used uninitialized in this function
	 */
	req_ats = NULL;

	if (ctx->schema == NULL) {
		ntg_message(ctx, nschema_msg);
		return -1;
	}

	e = 0;

	/*
	 * Step 1:
	 * Get OC selections from User.
	 */
	r_ocs = i_ocs = NULL;
	c_abs = c_aux = c_str = 0;

	while (ctx->schema->ocs->abs[c_abs] != NULL)
		c_abs++;
	ocs_abs = nt_calloc(c_abs + 1, sizeof(char *));
	for (i = 0; i < c_abs; i++)
		ocs_abs[i] = ctx->schema->ocs->abs[i]->oc_names[0];

	while (ctx->schema->ocs->aux[c_aux] != NULL)
		c_aux++;
	ocs_aux = nt_calloc(c_aux + 1, sizeof(char *));
	for (i = 0; i < c_aux; i++)
		ocs_aux[i] = ctx->schema->ocs->aux[i]->oc_names[0];

	while (ctx->schema->ocs->str[c_str] != NULL)
		c_str++;
	ocs_str = nt_calloc(c_str + 1, sizeof(char *));
	for (i = 0; i < c_str; i++)
		ocs_str[i] = ctx->schema->ocs->str[i]->oc_names[0];

	ocs_abs[c_abs] = ocs_aux[c_aux] = ocs_str[c_str] = NULL;

	qsort(ocs_abs, c_abs, sizeof(char *), ntu_strcasecmp);
	qsort(ocs_aux, c_aux, sizeof(char *), ntu_strcasecmp);
	qsort(ocs_str, c_str, sizeof(char *), ntu_strcasecmp);

	r_ocs = nt_calloc(c_abs + c_aux + c_str + 1, sizeof(char *));
	r_ocs[0] = NULL;
	r_count = 0;

#if 0
	if ((choices = ntg_select_multi(ctx, " Choose Zero Or More [Abstract] ObjectClasses: ", (const char **)ocs_abs)) == NULL) {
		e = 1;
		goto ntg_newobject_oc_out;
	}
	for (i = 0; i < c_abs; i++)
		if (choices[i] == 1)
			r_ocs[r_count++] = ocs_abs[i];
	free(choices);
#endif

	if ((choices = ntg_select_multi(ctx, " Choose One Or More [Structural] ObjectClasses: ", (const char **)ocs_str)) == NULL) {
		e = 1;
		goto ntg_newobject_oc_out;
	}
	flag = 0;
	for (i = 0; i < c_str; i++)
		if (choices[i] == 1) {
			r_ocs[r_count++] = ocs_str[i];
			flag = 1;
		}
	free(choices);
	if (!flag) {
		ntg_message(ctx, nval_msg);
		e = -1;
		goto ntg_newobject_oc_out;
	}

	if ((choices = ntg_select_multi(ctx, " Choose Zero Or More [Auxiliary] ObjectClasses: ", (const char **)ocs_aux)) == NULL) {
		e = 1;
		goto ntg_newobject_oc_out;
	}
	for (i = 0; i < c_aux; i++)
		if (choices[i] == 1)
			r_ocs[r_count++] = ocs_aux[i];
	free(choices);

	r_ocs[r_count] = NULL;

	/*
	 * Step 2:
	 * Interpolate OCs and get values for required attributes from User.
	 */
	req_ats = NULL;

	if ((i_ocs = ntls_interpolate_ocs((const NTLS_SCHEMA *)ctx->schema, (const char **)r_ocs)) == NULL)
		ntg_error_va(ctx, "Unable To Interpolate ObjectClasses.");
	else if ((req_ats = ntls_acc_ats((const NTLS_SCHEMA *)ctx->schema, (const char **)i_ocs, 'r')) == NULL)
		ntg_error_va(ctx, "Unable To Get Required AttributeTypes.");
	else {
		NTL_ATTR *new;
		char *rdn, *tmp, *inp;
		size_t ats_count, ocs_count;
		char *newbasedn;
		char *newdn;

		ats_count = 0;
		while (req_ats[ats_count] != NULL)
			ats_count++;
		ocs_count = 0;
		while (i_ocs[ocs_count] != NULL)
			ocs_count++;

		new = nt_calloc(ats_count + ocs_count + 2 + 1, sizeof(NTL_ATTR));
		new[0].name = NULL;
		new[0].val = NULL;
		new[0].tsize = ats_count + ocs_count + 2 + 1;

		if (*(newbasedn = ntg_input(ctx, "Base DN (Return For Default): ", basedn)) == '\0') {
			free(newbasedn);
			e = -2;
			goto ntg_newobject_oc_out;
		} else if (*(rdn = ntg_input(ctx, "RDN Attribute Type (i.e. \"uid\" or \"cn\"): ", NULL)) == '\0') {
			free(newbasedn);
			free(rdn);
			e = -2;
			goto ntg_newobject_oc_out;
		} else if (*(inp = ntg_input(ctx, "Value for RDN Attribute (i.e. \"user\" or \"First Last\"): ", NULL)) == '\0') {
			free(newbasedn);
			free(rdn);
			free(inp);
			e = -2;
			goto ntg_newobject_oc_out;
		}

		for (i = 0; i_ocs[i] != NULL; i++)
			ntl_attr_add(new, "objectClass", i_ocs[i]);

		ntl_attr_add(new, rdn, inp);
		nt_asprintf(&newdn, "%s=%s,%s", rdn, inp, newbasedn);
		free(inp);
		free(newbasedn);

		for (i = 0; req_ats[i] != NULL; i++) {
			if (strcmp(req_ats[i], "objectClass") == 0)
				continue; /* don't ask for OC, we got 'em */
			else if (strcmp(req_ats[i], rdn) == 0)
				continue; /* don't ask for RDN attribute either */
			ntg_attribute_tip(ctx, req_ats[i]);
			nt_asprintf(&tmp, "Enter Value For \"%s\": ", req_ats[i]);
			inp = ntg_input(ctx, tmp, NULL);
			ntg_tip_down(ctx);
			free(tmp);
			if (*inp != '\0')
				ntl_attr_add(new, req_ats[i], inp);
			else {
				free(inp);
				ntg_message(ctx, nval_msg);
				e = 1;
				break;
			}
			free(inp);
		}
		free(rdn);

		/*
		 * Step 3:
		 * Actually create Object on LDAP Server.
		 */
		ntg_info_up(ctx, cr_msg);
		if ((e = ntl_add(ctx->ld, newdn, new)) != 0)
			ntg_error_ldap(ctx, "Creation Failed", newdn, e);
		ntg_info_down(ctx);

		free(newdn);
		ntl_entry_free(new);
	}

ntg_newobject_oc_out:
	free(ocs_abs);
	free(ocs_str);
	free(ocs_aux);
	free(r_ocs);
	if (i_ocs != NULL)
		ntls_ocs_free(i_ocs);
	if (req_ats != NULL)
		ntls_acc_free(req_ats);

	return e;
}

MENU *
ntg_menu(NTG_CTX *ctx, ITEM *items[])
{
	MENU *menu;

	menu = new_menu(items);
	(void)set_menu_win(menu, ctx->mainwin);
	(void)set_menu_mark(menu, "=>");
	(void)set_menu_format(menu, ctx->mainwin_maxy - 1, 1);

	(void)cbreak();
	(void)curs_set(0);
	(void)keypad(stdscr, TRUE);

	return menu;
}

ITEM *
ntg_menu_driver(NTG_CTX *ctx, MENU *menu, int type)
{
	ITEM **items;
	const char *p;
	char *s;
	int c;
	int n, i;

	ntg_refresh(ctx);
	for (;;) {
		(void)cbreak();
		(void)noecho();
		ntg_tip_down(ctx);
		switch (c = getch()) {
		case KEY_DOWN:
			(void)menu_driver(menu, REQ_DOWN_ITEM);
			break;
		case KEY_UP:
			(void)menu_driver(menu, REQ_UP_ITEM);
			break;
		case KEY_LEFT:
			(void)menu_driver(menu, REQ_LEFT_ITEM);
			break;
		case KEY_RIGHT:
			(void)menu_driver(menu, REQ_RIGHT_ITEM);
			break;
		case KEY_NPAGE:
			if (menu_driver(menu, REQ_SCR_DPAGE) != E_OK)
				(void)menu_driver(menu, REQ_LAST_ITEM);
			break;
		case KEY_PPAGE:
			if (menu_driver(menu, REQ_SCR_UPAGE) != E_OK)
				(void)menu_driver(menu, REQ_FIRST_ITEM);
			break;
		case KEY_HOME:
			(void)menu_driver(menu, REQ_FIRST_ITEM);
			break;
		case KEY_END:
			(void)menu_driver(menu, REQ_LAST_ITEM);
			break;
		case 0x207: /* Shift+Up */
			(void)menu_driver(menu, REQ_SCR_ULINE);
			break;
		case 0x200: /* Shift+Down */
			(void)menu_driver(menu, REQ_SCR_DLINE);
			break;
		case ' ':
			(void)menu_driver(menu, REQ_TOGGLE_ITEM);
			break;
		case 0x0: /* Ctrl+Space */
			items = menu_items(menu);
			n = item_count(menu);
			for (i = 1; i < n && items[i] != NULL; i++) {
				if (item_value(items[i]) == TRUE)
					(void)set_item_value(items[i], FALSE);
				else
					(void)set_item_value(items[i], TRUE);
			}
			break;
		case 0x203: /* Ctrl+Left */
			items = menu_items(menu);
			n = item_count(menu);
			for (i = 1; i < n && items[i] != NULL; i++)
				(void)set_item_value(items[i], FALSE);
			break;
		case 0x205: /* Ctrl+Right */
			items = menu_items(menu);
			n = item_count(menu);
			for (i = 1; i < n && items[i] != NULL; i++)
				(void)set_item_value(items[i], TRUE);
			break;
		case 0x208: /* Ctrl+Up */
			if (item_value(current_item(menu)) == TRUE)
				(void)set_item_value(current_item(menu), FALSE);
			else
				(void)set_item_value(current_item(menu), TRUE);
			(void)menu_driver(menu, REQ_UP_ITEM);
			break;
		case 0x201: /* Ctrl+Down */
			if (item_value(current_item(menu)) == TRUE)
				(void)set_item_value(current_item(menu), FALSE);
			else
				(void)set_item_value(current_item(menu), TRUE);
			(void)menu_driver(menu, REQ_DOWN_ITEM);
			break;
		case KEY_ESC:
			return NULL;
		case 0x9b: /* Ctrl+Esc */
			c = ntg_cdkmenu_display(ctx, ctx->cdkmenus[type == NTG_MENU_SIMPLE ? 0 : 1]);
			if (c == 0)
				break;
			else if (c == NTG_MENU_RET_RETURN) {
				ctx->tmpreturn = 0;
				return NULL;
			} else if (c < 0) {
				ctx->tmpreturn = c;
				return NULL;
			}
			/* FALLTHROUGH */
		default:
			if (c == ctx->keys->search || c == ctx->keys->search_dn) {
				int flag;

				if (c == ctx->keys->search)
					s = ntg_input(ctx, "Plaintext Search Left Side For: ", NULL);
				else
					s = ntg_input(ctx, "Plaintext Search DN For: ", NULL);
				if (*s == '\0') {
					free(s);
					break;
				}

				/*
				 * Start search at current item.  If not found, wrap around.
				 * If possible, advance at least one position.
				 */
				i = 0;
				n = item_count(menu);
				items = menu_items(menu);
				while (items[i] != NULL && items[i] != current_item(menu))
					i++;
				flag = 0;
				if (items[i] != NULL) {
					i++;
					for (; i < n; i++)
						if (c == ctx->keys->search) {
							if (strcasestr(item_name(items[i]), s) != NULL) {
								(void)set_current_item(menu, items[i]);
								flag = 1;
								break;
							}
						} else {
							if (strcasestr(item_description(items[i]), s) != NULL) {
								(void)set_current_item(menu, items[i]);
								flag = 1;
								break;
							}
						}
				}
				if (items[i] == NULL || !flag)
					for (i = 1; i < n; i++) {
						if (c == ctx->keys->search) {
							if (strcasestr(item_name(items[i]), s) != NULL) {
								(void)set_current_item(menu, items[i]);
								break;
							}
						} else {
							if (strcasestr(item_description(items[i]), s) != NULL) {
								(void)set_current_item(menu, items[i]);
								break;
							}
						}
					}
				free(s);
				/* workaround (n)curses Bug/Feature */
				unpost_menu(menu);
				post_menu(menu);
				break;
			} else if (c == ctx->keys->suicide) {
				ctx->tmpreturn = NTG_MENU_RET_SUICIDE;
				return NULL;
			} else if (type == NTG_MENU_SIMPLE) {
				if (c == '\n')
					return current_item(menu);
				break;
			}

			if (c == ctx->keys->cr_tpl) {
				ctx->tmpreturn = NTG_MENU_RET_CREATE_FROM_TPL;
				return NULL;
			} else if (c == ctx->keys->cr_oc) {
				ctx->tmpreturn = NTG_MENU_RET_CREATE_FROM_OC;
				return NULL;
			} else if (c == ctx->keys->refresh) {
				ctx->tmpreturn = NTG_MENU_RET_REFRESH;
				return NULL;
			} else if (c == ctx->keys->del) {
				ctx->tmpreturn = NTG_MENU_RET_DEL;
				return NULL;
			} else if (c == ctx->keys->del_many) {
				ctx->tmpreturn = NTG_MENU_RET_DELM;
				return NULL;
			} else if (c == ctx->keys->display_auto) {
				ctx->tmpflags = 1;
				return current_item(menu);
			} else if (c == ctx->keys->display_tpl) {
				ctx->tmpflags = 0;
				return current_item(menu);
			} else if (c == ctx->keys->dive) {
				ctx->tmpreturn = NTG_MENU_RET_DIVE;
				return NULL;
			} else if (c == ctx->keys->ret) {
				ctx->tmpreturn = 0;
				return NULL;
			} else if (c == ctx->keys->refsearch) {
				ctx->tmpreturn = NTG_MENU_RET_REFSEARCH;
				return NULL;
			} else if (c == ctx->keys->subsearch) {
				ctx->tmpreturn = NTG_MENU_RET_SUBSEARCH;
				return NULL;
			} else if (c == ctx->keys->chg_base) {
				ctx->tmpreturn = NTG_MENU_RET_CHBASE;
				return NULL;
			} else if (c == ctx->keys->chg_sort) {
				ctx->tmpreturn = NTG_MENU_RET_CHSORT;
				return NULL;
			} else if (c == ctx->keys->tgl_operational) {
				ctx->cfg->showop = !ctx->cfg->showop;
				ctx->tmpreturn = NTG_MENU_RET_REFRESH;
				return NULL;
			} else if (c == ctx->keys->scope) {
				ctx->tmpreturn = NTG_MENU_RET_SCOPE;
				return NULL;
			} else if (c == ctx->keys->ldif) {
				ctx->tmpreturn = NTG_MENU_RET_LDIF;
				return NULL;
			} else if (c == ctx->keys->copy) {
				ctx->tmpreturn = NTG_MENU_RET_COPY;
				return NULL;
			} else if (c == ctx->keys->copy_many) {
				ctx->tmpreturn = NTG_MENU_RET_COPYM;
			} else if (c == ctx->keys->move) {
				ctx->tmpreturn = NTG_MENU_RET_MOVE;
				return NULL;
			} else if (c == ctx->keys->move_many) {
				ctx->tmpreturn = NTG_MENU_RET_MOVEM;
				return NULL;
			} else {
				/*
				 * real "default" case
				 */
				if ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z')) {
					char *sp;
					int flag;

					/*
					 * Start search at current item, if not found wrap around.
					 * If possible, advance at least one position.
					 *
					 * Allow search in first part of left side (via a-z)
					 * and in second part (via A-Z).
					 */
					i = 0;
					n = item_count(menu);
					items = menu_items(menu);
					while (items[i] != NULL && items[i] != current_item(menu))
						i++;
					flag = 0;
					if (items[i] != NULL) {
						i++;
						for (; i < n; i++) {
							p = item_name(items[i]);
							sp = strchr(p, ':') + 2;
							if (c >= 'A' && c <= 'Z')
								sp = strchr(sp, ':') + 2;
							if (toupper((int)*sp) == toupper(c)) {
								(void)set_current_item(menu, items[i]);
								flag = 1;
								break;
							}
						}
					}
					if (items[i] == NULL || !flag)
						for (i = 1; i < n; i++) {
							p = item_name(items[i]);
							sp = strchr(p, ':') + 2;
							if (c >= 'A' && c <= 'Z')
								sp = strchr(sp, ':') + 2;
							if (toupper((int)*sp) == toupper(c)) {
								(void)set_current_item(menu, items[i]);
								break;
							}
						}
				}
			}
		}
		(void)wrefresh(menu_win(menu));
	}
	/* NOTREACHED */
}

void
ntg_subtitle(NTG_CTX *ctx, const char *s)
{
	size_t width;

	width = getmaxx(ctx->topwin);
	(void)mvwhline(ctx->topwin, 1, 0, ' ', width);

	if (s != NULL) {
		(void)snprintf(ctx->subtitle, sizeof(ctx->subtitle), "%s", s);
		/* truncate if string is longer than a/v screenwidth */
		if (strlen(ctx->subtitle) > width - 4)
			ctx->subtitle[width - 4] = '\0';
		(void)wattrset(ctx->topwin, C_STRONG);
		(void)mvwprintw(ctx->topwin, 1, 1, "(%s)", ctx->subtitle);
	}
}

void
ntg_title(NTG_CTX *ctx, const char *s)
{
	int width;

	width = getmaxx(ctx->topwin);

	ntg_subtitle(ctx, NULL);

	(void)wattrset(ctx->topwin, A_NORMAL | C_TEXT);
	(void)mvwhline(ctx->topwin, 0, 0, 0, width);

	(void)snprintf(ctx->title, sizeof(ctx->title), "%s", s);
	/* truncate if string is longer than a/v screenwidth */
	if (strlen(ctx->title) > (size_t)width - 6)
		ctx->title[width - 6] = '\0';
	(void)wattrset(ctx->topwin, C_EM);
	(void)mvwprintw(ctx->topwin, 0, 1, "[%s]", ctx->title);
	(void)wattrset(ctx->topwin, C_TEXT);
	if ((int)(width - strlen(myname) - 2 - strlen(release) - 3) > 0)
		(void)mvwprintw(ctx->topwin, 0, (int)(width - strlen(myname) - 2 - strlen(release) - 3), "%s v%s", myname, release);
}

void
ntg_refresh(NTG_CTX *ctx)
{
	refreshCDKScreen(ctx->cdkscreen);
	(void)refresh();
	(void)touchwin(ctx->mainwin);
	(void)wrefresh(ctx->mainwin);
	(void)touchwin(ctx->topwin);
	(void)wrefresh(ctx->topwin);
	(void)touchwin(ctx->bottomwin);
	(void)wrefresh(ctx->bottomwin);
}

ITEM **
ntg_items(const char **name, const char **desc)
{
	ITEM **items;
	int c;
	int i;

	if (*name == NULL || *desc == NULL)
		return NULL;

	/* count amount */
	c = 0;
	while (name[c] != NULL && desc[c] != NULL)
		c++;

	/* fill */
	items = nt_calloc((size_t)c + 1, sizeof(ITEM *));
	for (i = 0; i < c; i++)
		items[i] = new_item(name[i], desc[i]);
	items[i] = NULL;

	return items;
}

void
ntg_items_free(ITEM **items)
{
	int i;

	if (items == NULL)
		return;

	for (i = 0; items[i] != NULL; i++)
		free_item(items[i]);
	free(items);
	items = NULL;
}

void
ntg_set_edit(FORM *form, int on)
{
	FIELD **fields;
	int count;
	int i, j;

	fields = form_fields(form);
	count = field_count(form);

	if (on) {
		/*
		 * workaround (n)curses Bug/Feature
		 */
		for (j = 0; j < 2; j++) {
			if (j == 0)
				(void)form_driver(form, REQ_NEXT_FIELD);
			else
				(void)form_driver(form, REQ_PREV_FIELD);

			for (i = 0; i < count; i++) {
				(void)field_opts_on(fields[i], O_EDIT);
				if (field_opts(fields[i]) & O_ACTIVE) {
					(void)field_opts_on(fields[i], O_EDIT);
					if (!(field_back(fields[i]) & (C_STRONG|C_TEXT|C_EM)))
						(void)set_field_back(fields[i], (field_back(fields[i]) & ~A_NORMAL) | A_REVERSE | C_BLUE);
					else
						(void)set_field_back(fields[i], (field_back(fields[i]) & ~A_NORMAL) | A_REVERSE);
				}
			}
		}
	} else {
		for (j = 0; j < 2; j++) {
			if (j == 0)
				(void)form_driver(form, REQ_NEXT_FIELD);
			else
				(void)form_driver(form, REQ_PREV_FIELD);

			for (i = 0; i < count; i++) {
				(void)field_opts_off(fields[i], O_EDIT);
				if (field_opts(fields[i]) & O_ACTIVE) {
					(void)field_opts_off(fields[i], O_EDIT);
					if (field_back(fields[i]) & C_BLUE & ~(C_STRONG|C_TEXT|C_EM))
						(void)set_field_back(fields[i], (field_back(fields[i]) & ~(A_REVERSE|C_BLUE)) | A_NORMAL);
					else
						(void)set_field_back(fields[i], (field_back(fields[i]) & ~A_REVERSE) | A_NORMAL);
				}
			}
		}
	}
}

void
ntg_set_virgin(FORM *form)
{
	FIELD **fields;
	int count;
	int i;

	fields = form_fields(form);
	count = field_count(form);

	for (i = 0; i < count; i++)
		(void)set_field_status(fields[i], FALSE);
}

int
ntg_form_driver(NTG_CTX *ctx, FORM *form)
{
	FIELD **f;
	char s[12];
	char *p, *tmp, *tmp2;
	int fieldcount;
	int edit, showtip, changed;
	int px, py;
	int c;
	int r;
	int i;

	f = form_fields(form);
	fieldcount = field_count(form);

	edit = showtip = 0;

	ntg_set_edit(form, edit);
	ntg_set_virgin(form);

	for (;;) {
		if (showtip)
			ntg_attribute_tip(ctx, field_userptr(current_field(form)));

		changed = 0;
		for (i = 0; i < fieldcount; i++)
			if (field_status(f[i]) == TRUE) {
				set_field_pad(f[i], '<');
				changed++;
			}

		px = getcurx(ctx->mainwin);
		py = getcury(ctx->mainwin);

		(void)wmove(ctx->mainwin, ctx->mainwin_maxy - 1, 0);
		(void)whline(ctx->mainwin, ' ', ctx->mainwin_maxx);

		(void)snprintf(s, sizeof(s), "[Page %d/%d]", form_page(form) + 1, form->maxpage);
		(void)mvwprintw(ctx->mainwin, ctx->mainwin_maxy - 1, ctx->mainwin_maxx - (int)sizeof(s), "%s", s);

		if (changed)
			(void)mvwprintw(ctx->mainwin, ctx->mainwin_maxy - 1, 0, "*** %d values are unsaved ***", changed);

		p = ntl_strip(field_buffer(current_field(form), 1));
		if (*p != '\0')
			(void)mvwprintw(ctx->mainwin, ctx->mainwin_maxy - 1, 34, "[current field type: %s]", p);
		free(p);

		(void)wmove(ctx->mainwin, py, px);
		(void)wrefresh(form_sub(form));

		(void)cbreak();
		(void)noecho();

		switch (c = getch()) {
		case KEY_ESC:
			return 0;
		case 0x0e: /* ^N */
			return NTG_PAGER_RET_NEXT;
		case 0x10: /* ^P */
			return NTG_PAGER_RET_PREV;
		default:
			if (c == ctx->keys->ret)
				return 0;
			else if (c == ctx->keys->write) {
				if (form_driver(form, REQ_VALIDATION) != E_OK) {
					ntg_message(ctx, fval_msg);
					/* workaround (n)curses Bug/Feature */
					(void)unpost_form(form);
					(void)post_form(form);
				} else
					return NTG_PAGER_RET_WRITE;
			} else if (c == ctx->keys->del)
				return NTG_PAGER_RET_DEL;
			else if (c == ctx->keys->cr_attr)
				return NTG_PAGER_RET_CREATE_ATTR;
			else if (c == ctx->keys->cr_tpl)
				return NTG_PAGER_RET_CREATE_FROM_TPL;
			else if (c == ctx->keys->cr_oc)
				return NTG_PAGER_RET_CREATE_FROM_OC;
			else if (c == ctx->keys->ldif)
				return NTG_PAGER_RET_LDIF;
			else if (c == ctx->keys->refresh)
				return NTG_PAGER_RET_RELOAD;
			else if (c == ctx->keys->edit) {
				edit = !edit;
				ntg_set_edit(form, edit);
			} else if (c == '\n') {
				/*
				 * Hack to allow artificial \n represented by a ''.
				 * See ntl.c ntl_in_nl/ntl_out_nl also.
				*/
				if (edit) {
					(void)form_driver(form, REQ_INS_CHAR);
					(void)waddstr(form->w, "");
					(void)form_driver(form, REQ_RIGHT_CHAR);
				}
			} else if (c == ctx->keys->tip) {
				if (showtip) {
					showtip = 0;
					ntg_tip_down(ctx);
				} else
					showtip = 1;
			} else if (c == ctx->keys->subsearch)
				return NTG_PAGER_RET_SEARCH;
			else if (c == ctx->keys->password)
				return NTG_PAGER_RET_PASSWD;
			else if (c == ctx->keys->modrdn)
				return NTG_PAGER_RET_MODRDN;
			else if (c == ctx->keys->copy)
				return NTG_PAGER_RET_COPY;
			else if (c == ctx->keys->move)
				return NTG_PAGER_RET_MOVE;
			else if (c == ctx->keys->suicide)
				return NTG_PAGER_RET_SUICIDE;
			else if (c == ctx->keys->open_editor) {
				tmp2 = ntl_strip(field_buffer(current_field(form), 0));
				ntl_in_nl(tmp2);
				if ((tmp = ntg_open_editor(ctx, tmp2)) != NULL) {
					if (strcmp(tmp, tmp2) != 0) {
						ntl_out_nl(tmp);
						(void)set_field_buffer(current_field(form), 0, tmp);
						(void)set_field_status(current_field(form), TRUE);
					}
					free(tmp);
				}
				free(tmp2);
			} else if (c == ctx->keys->view_attr) {
				tmp = ntl_strip(field_buffer(current_field(form), 0));
				ntl_in_nl(tmp);
				ntg_display_va(ctx, "Attribute: %s\n%s\n", (char *)field_userptr(current_field(form)), tmp);
				free(tmp);
			} else if (c == ctx->keys->validate) {
				r = form_driver(form, REQ_VALIDATION);
				if (r == E_INVALID_FIELD) {
					ntg_message(ctx, fval_msg);
					/* workaround (n)curses Bug/Feature */
					(void)unpost_form(form);
					(void)post_form(form);
				}
			} else {
				/*
				 * real "default" case
				 */
				r = ntg_form_driver_super(form, c);
				if (r == E_INVALID_FIELD) {
					ntg_message(ctx, fval_msg);
					/* workaround (n)curses Bug/Feature */
					(void)unpost_form(form);
					(void)post_form(form);
				}
			}
		}
	}
	/* NOTREACHED */
}

int
ntg_form_driver_super(FORM *form, int c)
{

	switch (c) {
	case KEY_DOWN:
	case 0x09: /* Tab */
		return form_driver(form, REQ_NEXT_FIELD);
	case KEY_UP:
	case 0x161: /* Shift+Tab */
		return form_driver(form, REQ_PREV_FIELD);
	case 0x189: /* Shift+Left */
		return form_driver(form, REQ_PREV_WORD);
	case 0x192: /* Shift+Right */
		return form_driver(form, REQ_NEXT_WORD);
	case 0x207: /* Shift+Up */
		return form_driver(form, REQ_PREV_CHOICE);
	case 0x200: /* Shift+Down */
		return form_driver(form, REQ_NEXT_CHOICE);
	case KEY_LEFT:
		return form_driver(form, REQ_LEFT_CHAR);
	case KEY_RIGHT:
		return form_driver(form, REQ_RIGHT_CHAR);
	case KEY_NPAGE:
		if (form_page(form) + 1 >= form->maxpage)
			return form_driver(form, REQ_LAST_FIELD);
		return form_driver(form, REQ_NEXT_PAGE);
	case KEY_PPAGE:
		if (form_page(form) <= 0)
			return form_driver(form, REQ_FIRST_FIELD);
		return form_driver(form, REQ_PREV_PAGE);
	case 0x17: /* ^W */
		return form_driver(form, REQ_DEL_WORD);
	case KEY_BACKSPACE:
	case 0x08: /* ^H */
		return form_driver(form, REQ_DEL_PREV);
	case KEY_IC: /* Ins */
		return (form_driver(form, REQ_INS_MODE) ||
			form_driver(form, REQ_OVL_MODE));
	case KEY_DC: /* Del */
		return form_driver(form, REQ_DEL_CHAR);
	case 0x04: /* ^D */
	case 0x17f: /* Shift+Del */
		return form_driver(form, REQ_CLR_EOL);
	case 0x15: /* ^U */
	case 0x0b: /* ^K */
		return form_driver(form, REQ_DEL_LINE);
	case KEY_HOME:
	case 0x01: /* ^A */
		return form_driver(form, REQ_BEG_LINE);
	case KEY_END:
	case 0x05: /* ^E */
		return form_driver(form, REQ_END_LINE);
	case 0x187: /* Shift+Home */
	case 0x02: /* ^B */
		return form_driver(form, REQ_FIRST_FIELD);
	case 0x182: /* Shift+End */
		return form_driver(form, REQ_LAST_FIELD);
	default:
		return form_driver(form, c);
	}
	/* NOTREACHED */
}

int
ntg_open_shell(NTG_CTX *ctx, const char *cmd)
{
	WINDOW *win_stdscr, *win_rootwin;
	int r;

	win_stdscr = dupwin(stdscr);
	win_rootwin = dupwin(ctx->rootwin);

	(void)wclear(stdscr);
	(void)wrefresh(stdscr);
	(void)savetty();
	(void)reset_shell_mode();

	r = system(cmd);

	(void)resetty();
	(void)wclear(stdscr);
	(void)wrefresh(stdscr);

	(void)overwrite(win_stdscr, stdscr);
	(void)overwrite(win_rootwin, ctx->rootwin);

	(void)delwin(win_stdscr);
	(void)delwin(win_rootwin);

	ntg_refresh(ctx);

	return r;
}

char *
ntg_open_editor(NTG_CTX *ctx, const char *buf)
{
	FILE *fp;
	size_t len;
	char tmp[64], cmd[128], outbuf[65536];
	char *out;
	int fd;
	int r;

	*tmp = *cmd = *outbuf = '\0';
	out = NULL;

	ntg_info_up(ctx, editor_msg);

	(void)strncpy(tmp, "/tmp/ntg.XXXXXX", sizeof(tmp) - 1);
	tmp[sizeof(tmp) - 1] = '\0';
	fd = -1;
	if ((fd = mkstemp(tmp)) == -1 || (fp = fdopen(fd, "w+")) == NULL) {
		if (fd != -1) {
			(void)unlink(tmp);
			(void)close(fd);
		}
		ntg_error_va(ctx, "Unable To Open Temporary File:\n%s", tmp);
		goto ntg_open_editor_out2;
	}

	len = strlen(buf);
	if (fwrite(buf, len, 1, fp) != 1 && len != 0) {
		ntg_error_va(ctx, "Unable To Write To Temporary File:\n%s", tmp);
		goto ntg_open_editor_out;
	}
	(void)fsync(fd);
	(void)fclose(fp);

	(void)snprintf(cmd, sizeof(cmd), "%s '%s'", getenv("EDITOR") == NULL ? "vi" : getenv("EDITOR"), tmp);
	r = ntg_open_shell(ctx, cmd);

	if (r == -1)
		ntg_error_va(ctx, "Unable To Launch Editor:\n%s", cmd);
	else if (r != 0)
		ntg_error_va(ctx, "Editor Returned Non-Zero Error Code Indicating Error Or Abort");
	else {
		/* success */
		if ((fp = fopen(tmp, "r")) == NULL)
			ntg_error_va(ctx, "Unable To Read-Back Temporary File:\n%s", tmp);
		else {
			len = 0;
			while (len < sizeof(outbuf) && fread(outbuf + len, 1, 1, fp) == 1)
				len++;
			outbuf[len] = '\0';

			if (len > 0) {
				out = nt_calloc(len + 1, sizeof(char));
				(void)memcpy(out, outbuf, len);
				out[len] = '\0';
				/* strip trailing newline */
				if (out[len - 1] == '\n')
					out[len - 1] = '\0';
			} else {
				out = nt_calloc(1, sizeof(char));
				*out = '\0';
			}
			(void)fclose(fp);
		}
	}

ntg_open_editor_out:
	(void)unlink(tmp);
ntg_open_editor_out2:
	ntg_info_down(ctx);

	return out;
}

static int
ntg_input_keybind(EObjectType cdkType, void *object, void *clientData, chtype key)
{
	CDKENTRY *entry;

	/* unused */
	(void)clientData;
	(void)cdkType;

	entry = object;

	if (key == KEY_HOME || key == 0x1)
		(void)injectCDKEntry(entry, (chtype)CDK_BEGOFLINE);
	else if (key == KEY_END || key == 0x5)
		(void)injectCDKEntry(entry, (chtype)CDK_ENDOFLINE);
	else if (key == KEY_DC)
		(void)injectCDKEntry(entry, (chtype)DELETE);
	else if (key == 0xb) /* ^K */
		(void)injectCDKEntry(entry, (chtype)CDK_ERASE);
	else if (key == KEY_BACKSPACE) {
		(void)injectCDKEntry(entry, (chtype)KEY_LEFT);
		(void)injectCDKEntry(entry, (chtype)DELETE);
	} else
		return 1;

	return 0;
}

/**
 * Read line from user, returns malloc()ed string and "" on error or empty
 */
char *
ntg_input(NTG_CTX *ctx, const char *desc, const char *dflt)
{
	CDKENTRY *input;
	char *s, *p;
	int curs_old;

	if (ctx->cdklabel != NULL) {
		eraseCDKLabel(ctx->cdklabel);
		refreshCDKScreen(ctx->cdkscreen);
	}

	(void)refresh();

	if ((input = newCDKEntry(ctx->cdkscreen, CENTER, CENTER, cdk_input_str, (char *)desc, A_BOLD, (unsigned long) '_', vMIXED, 60, 0, 512, TRUE, FALSE)) == NULL)
		ntg_fail(ctx, "unexpected result from CDKEntry", NTG_E_INPUT);
	setCDKEntryBackgroundColor(input, cdk_input_bgcolor);
	setCDKEntryValue(input, (char *)dflt);
	if (CDK_VERSION_MAJOR == "4") {
		(void)injectCDKEntry(input, (chtype)CDK_BEGOFLINE);
		setCDKEntryPreProcess(input, ntg_input_keybind, NULL);
	}

	curs_old = curs_set(1);
	p = activateCDKEntry(input, NULL);
	if (p != NULL)
		s = nt_strdup(p);
	else
		s = nt_strdup("");
	eraseCDKEntry(input);
	destroyCDKEntry(input);
	refreshCDKScreen(ctx->cdkscreen);
	(void)curs_set(curs_old);

	if (ctx->cdklabel != NULL)
		activateCDKLabel(ctx->cdklabel, NULL);

	return s;
}

void
ntg_popup(NTG_CTX *ctx, const char **msg, const char *color)
{
	CDKLABEL *popup;
	int curs_old;
	int size;

	size = 0;
	while (msg[size] != NULL)
		size++;

	if (ctx->cdklabel != NULL) {
		eraseCDKLabel(ctx->cdklabel);
		refreshCDKScreen(ctx->cdkscreen);
	}

	(void)refresh();

	curs_old = curs_set(0);
	popup = newCDKLabel(ctx->cdkscreen, CENTER, CENTER, (char **)msg, size, TRUE, FALSE);
	setCDKLabelBackgroundColor(popup, (char *)color);
	activateCDKLabel(popup, NULL);
	(void)getch();

	eraseCDKLabel(popup);
	destroyCDKLabel(popup);
	refreshCDKScreen(ctx->cdkscreen);
	(void)curs_set(curs_old);

	if (ctx->cdklabel != NULL)
		activateCDKLabel(ctx->cdklabel, NULL);
}

void
ntg_error_ldap(NTG_CTX *ctx, const char *msg, const char *dn, int e)
{
	const char *e_msg[4];
	char **ld_msg;

	if (e >= 0) {
		ld_msg = ntl_ldap_err(msg, dn, ctx->ld, e);
		ntg_error_va(ctx, "</U>%s<!U>\n * %s %s\n * %s %s\n * %s %s\n * %s %s",
		    ld_msg[0], ld_msg[1], ld_msg[2], ld_msg[3], ld_msg[4], ld_msg[5], ld_msg[6], ld_msg[7], ld_msg[8]);
		ntl_ldap_err_free(ld_msg);
	} else {
		e_msg[0] = (char *)msg;
		e_msg[1] = (char *)dn;
		e_msg[2] = ntl_error_str;
		e_msg[3] = NULL;
		ntg_error(ctx, e_msg);
	}
}

char **
ntg_strnl_to_strstr(const char *s)
{
	char buf[2048];
	char **lines;
	int i, j, count, last;

	count = 1;
	for (i = 0; s[i] != '\0'; i++)
		if (s[i] == '\n')
			count++;

	lines = nt_calloc((size_t)(count + 2), sizeof(char *));

	count = 0;
	last = 0;
	for (i = 0; s[i] != '\0'; i++)
		if (s[i] == '\n')
			if (i - last < (int)sizeof(buf) - 1) {
				for (j = 0; j < i - last; j++)
					buf[j] = s[last + j];
				buf[j] = '\0';
				nt_asprintf(&lines[count], "%s", buf);
				count++;
				last = i + 1;
			}
	for (j = 0; j < i - last; j++)
		buf[j] = s[last + j];
	buf[j] = '\0';
	nt_asprintf(&lines[count], "%s", buf);
	lines[++count] = NULL;

	return lines;
}

void
ntg_strstr_free(char **s)
{
	int i;

	if (s == NULL)
		return;

	for (i = 0; s[i] != NULL; i++)
		free(s[i]);
	free(s);
}

void
ntg_va(NTG_CTX *ctx, void fn(NTG_CTX *, const char **), const char *fmt, va_list ap)
{
	char buf[65536];
	char **lines;

	if (vsnprintf(buf, sizeof(buf), fmt, ap) < 0)
		return;
	va_end(ap);

	lines = ntg_strnl_to_strstr(buf);
	fn(ctx, (const char **)lines);
	ntg_strstr_free(lines);
}

void
ntg_error(NTG_CTX *ctx, const char **msg)
{
	ntg_popup(ctx, msg, "</58>");
}

void
ntg_error_va(NTG_CTX *ctx, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	ntg_va(ctx, ntg_error, fmt, ap);
}

void
ntg_success(NTG_CTX *ctx, const char **msg)
{
	ntg_popup(ctx, msg, "</59>");
}

void
ntg_success_va(NTG_CTX *ctx, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	ntg_va(ctx, ntg_success, fmt, ap);
}

void
ntg_message(NTG_CTX *ctx, const char **msg)
{
	ntg_popup(ctx, msg, "</57>");
}

void
ntg_message_va(NTG_CTX *ctx, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	ntg_va(ctx, ntg_message, fmt, ap);
}

void
ntg_display(NTG_CTX *ctx, const char **msg)
{
	CDKVIEWER *viewer;
	const char *button[] = { "</5><Close><!5>", NULL };
	char title[128];
	int h, v;
	int size;

	size = 0;
	while (msg[size] != NULL)
		size++;

	h = ctx->mainwin_maxx * 8/10;
	v = ctx->mainwin_maxy * 6/10;
	if (h < 50)
		h = 50;
	if (v < 10)
		v = 10;

	if ((viewer = newCDKViewer(ctx->cdkscreen, CENTER, CENTER, v, h, (char **)button, 1, A_REVERSE, TRUE, FALSE)) == NULL)
		return;

	(void)snprintf(title, sizeof(title), "<R></1> [ %s ] <!1>", msg[0]);
	setCDKViewer(viewer, title, (char **)msg + 1, size - 1, A_REVERSE, TRUE, TRUE, TRUE);
	setCDKViewerBackgroundColor(viewer, cdk_display_bgcolor);
	(void)activateCDKViewer(viewer, NULL);
	eraseCDKViewer(viewer);
	destroyCDKViewer(viewer);
	refreshCDKScreen(ctx->cdkscreen);
}

void
ntg_display_va(NTG_CTX *ctx, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	ntg_va(ctx, ntg_display, fmt, ap);
}

void
ntg_tip_up(NTG_CTX *ctx, const char **info)
{
	int size;

	size = 0;
	while (info[size] != NULL)
		size++;

	if (ctx->cdktip != NULL)
		ntg_tip_down(ctx);

	ctx->cdktip = newCDKLabel(ctx->cdkscreen, RIGHT, TOP, (char **)info, size, TRUE, FALSE);
	setCDKLabelBackgroundColor(ctx->cdktip, cdk_tip_bgcolor);
	activateCDKLabel(ctx->cdktip, NULL);
}

void
ntg_tip_down(NTG_CTX *ctx)
{

	if (ctx->cdktip == NULL)
		return;

	eraseCDKLabel(ctx->cdktip);
	destroyCDKLabel(ctx->cdktip);
	refreshCDKScreen(ctx->cdkscreen);
	ctx->cdktip = NULL;
}

void
ntg_attribute_tip(NTG_CTX *ctx, const char *attr)
{
	const LDAPAttributeType *at;
	const char *tip[20];

	ntg_tip_down(ctx);

	if (ctx->schema == NULL)
		return;

	if ((at = ntls_get_at_from_name((const LDAPAttributeType **) ctx->schema->ats, attr)) == NULL)
		return;

	tip[0] = "[</B>Info About Attribute (Name, Descr)<!B>]";
	tip[1] = at->at_names[0];
	tip[2] = at->at_desc == NULL ? "" : at->at_desc;
	tip[3] = at->at_obsolete == LDAP_SCHEMA_YES ? "Obsolete" : "";
	tip[4] = "";

	tip[5] = "[</B>Attribute Syntax (Descr, OID)<!B>]";
	tip[6] = "(unknown/not specified)";
	if (at->at_syntax_oid == NULL)
		; /* ignore */
	else if (strcmp(at->at_syntax_oid, "1.3.6.1.1.16.1") == 0)
		tip[6] = "UUID";
	else if (strcmp(at->at_syntax_oid, "1.3.6.1.4.1.1466.115.121.1.5") == 0)
		tip[6] = "Binary (BER/DER date)";
	else if (strcmp(at->at_syntax_oid, "1.3.6.1.4.1.1466.115.121.1.6") == 0)
		tip[6] = "Bit String (ex.: '0101111101'B) ";
	else if (strcmp(at->at_syntax_oid, "1.3.6.1.4.1.1466.115.121.1.7") == 0)
		tip[6] = "Boolean (boolean value, TRUE or FALSE)";
	else if (strcmp(at->at_syntax_oid, "1.3.6.1.4.1.1466.115.121.1.11") == 0)
		tip[6] = "Country String (two chars, ISO 3166; ex.: US, DE)";
	else if (strcmp(at->at_syntax_oid, "1.3.6.1.4.1.1466.115.121.1.12") == 0)
		tip[6] = "Distinguished Name (DN)";
	else if (strcmp(at->at_syntax_oid, "1.3.6.1.4.1.1466.115.121.1.15") == 0)
		tip[6] = "Directory String (UTF-8 string, ISO 10646)";
	else if (strcmp(at->at_syntax_oid, "1.3.6.1.4.1.1466.115.121.1.22") == 0)
		tip[6] = "Facsimile Telephone Number (faxnumber or faxnumber$params)";
	else if (strcmp(at->at_syntax_oid, "1.3.6.1.4.1.1466.115.121.1.24") == 0)
		tip[6] = "Generalized Time (X.208, ex.: 199412161032Z)";
	else if (strcmp(at->at_syntax_oid, "1.3.6.1.4.1.1466.115.121.1.26") == 0)
		tip[6] = "IA5String (ASCII string)";
	else if (strcmp(at->at_syntax_oid, "1.3.6.1.4.1.1466.115.121.1.27") == 0)
		tip[6] = "Integer (integer)";
	else if (strcmp(at->at_syntax_oid, "1.3.6.1.4.1.1466.115.121.1.28") == 0)
		tip[6] = "JPEG (JPEG picture in JFIF format)";
	else if (strcmp(at->at_syntax_oid, "1.3.6.1.4.1.1466.115.121.1.34") == 0)
		tip[6] = "Name And Optional UID (DN plus opt. UID in bitstring)";
	else if (strcmp(at->at_syntax_oid, "1.3.6.1.4.1.1466.115.121.1.36") == 0)
		tip[6] = "Numeric String (numeric string)";
	else if (strcmp(at->at_syntax_oid, "1.3.6.1.4.1.1466.115.121.1.38") == 0)
		tip[6] = "Object Identifier (OID)";
	else if (strcmp(at->at_syntax_oid, "1.3.6.1.4.1.1466.115.121.1.39") == 0)
		tip[6] = "Other Mailbox";
	else if (strcmp(at->at_syntax_oid, "1.3.6.1.4.1.1466.115.121.1.40") == 0)
		tip[6] = "Octet String (arbitrary octets)";
	else if (strcmp(at->at_syntax_oid, "1.3.6.1.4.1.1466.115.121.1.41") == 0)
		tip[6] = "Postal Address (ex.: 1234 Main St.$Anytown, CA 12345$USA)";
	else if (strcmp(at->at_syntax_oid, "1.3.6.1.4.1.1466.115.121.1.44") == 0)
		tip[6] = "Printable String (printable string)";
	else if (strcmp(at->at_syntax_oid, "1.3.6.1.4.1.1466.115.121.1.50") == 0)
		tip[6] = "Telephone Number (X.520; ex.: +1 512 305 0280)";
	else if (strcmp(at->at_syntax_oid, "1.3.6.1.4.1.1466.115.121.1.53") == 0)
		tip[6] = "UTC Time";
	tip[7] = at->at_syntax_oid == NULL ? "" : at->at_syntax_oid;

	tip[8] = "[</B>Qualifiers<!B>]";
	tip[9] = at->at_single_value == LDAP_SCHEMA_YES ? "Single" : "Multi";
	tip[10] = at->at_collective == LDAP_SCHEMA_YES ? "Collective" : "Non-Collective";
	tip[11] = at->at_no_user_mod == LDAP_SCHEMA_YES ? "Un-Changeable" : "Changeable";

	tip[12] = "[</B>Match Types (Equality, Ordering, Substr)<!B>]";
	tip[13] = at->at_equality_oid == NULL ? "" : at->at_equality_oid;
	tip[14] = at->at_ordering_oid == NULL ? "" : at->at_ordering_oid;
	tip[15] = at->at_substr_oid == NULL ? "" : at->at_substr_oid;

	tip[16] = "[</B>Usage<!B>]";
	switch(at->at_usage) {
	case LDAP_SCHEMA_USER_APPLICATIONS:
		tip[17] = "User Applications (the attribute type is non-operational)";
		break;
	case LDAP_SCHEMA_DIRECTORY_OPERATION:
		tip[17] = "Directory Operation (operational and pertinent to the directory itself)";
		break;
	case LDAP_SCHEMA_DISTRIBUTED_OPERATION:
		tip[17] = "Distributed Operation (operat. and pertinent to replication, shad., etc.)";
		break;
	case LDAP_SCHEMA_DSA_OPERATION:
		tip[17] = "DSA Operation (operat. and pertinent to the directory _server_ itself)";
		break;
	default:
		tip[17] = "(unknown usage)";
	}

	tip[18] = NULL;

	ntg_tip_up(ctx, tip);
}

void
ntg_info_up(NTG_CTX *ctx, const char **info)
{
	int size;

	size = 0;
	while (info[size] != NULL)
		size++;

	if (ctx->cdklabel != NULL)
		ntg_info_down(ctx);

	ctx->oldcursor = curs_set(0);
	ctx->cdklabel = newCDKLabel(ctx->cdkscreen, CENTER, CENTER, (char **)info, size, TRUE, FALSE);
	setCDKLabelBackgroundColor(ctx->cdklabel, cdk_info_bgcolor);
	activateCDKLabel(ctx->cdklabel, NULL);
}

void
ntg_info_down(NTG_CTX *ctx)
{

	if (ctx->cdklabel == NULL)
		return;

	eraseCDKLabel(ctx->cdklabel);
	destroyCDKLabel(ctx->cdklabel);
	refreshCDKScreen(ctx->cdkscreen);
	ctx->cdklabel = NULL;
	(void)curs_set(ctx->oldcursor);
}

int
ntg_yesno(NTG_CTX *ctx, const char **quest)
{
	CDKLABEL *q;
	int size;
	int c, r;

	size = 0;
	while (quest[size] != NULL)
		size++;

	if (ctx->cdklabel != NULL) {
		eraseCDKLabel(ctx->cdklabel);
		refreshCDKScreen(ctx->cdkscreen);
	}

	(void)use_default_colors();
	ctx->oldcursor = curs_set(0);
	q = newCDKLabel(ctx->cdkscreen, CENTER, CENTER, (char **)quest, size, TRUE, FALSE);
	setCDKLabelBackgroundColor(q, cdk_yesno_bgcolor);
	activateCDKLabel(q, NULL);

	c = ' ';
	while (c != 'y' && c != 'Y' && c != 'n' && c != 'N')
		c = waitCDKLabel(q, (char)0);
	/* default to no == 0 */
	r = 0;
	if (c == 'y' || c == 'Y')
		r = 1;

	eraseCDKLabel(q);
	destroyCDKLabel(q);
	(void)curs_set(ctx->oldcursor);

	if (ctx->cdklabel != NULL)
		activateCDKLabel(ctx->cdklabel, NULL);

	return r;
}

int
ntg_search(NTG_CTX *ctx, const char *base, int scope, const char *p_fltr, int (*displayfn)(NTG_CTX *, const char *, const char *, int))
{
	/*
	 * We use fltr on stack to avoid leak on DIVE and co., when
	 * using malloc()ed memory, since we need to pass current filter
	 * along the way.
	 *
	 * fltr points normally to fltr_buf, but may be NULL.
	 */
	LDAPMessage *res;
	NTL_ATTR *attr;
	char fltr_buf[1024];
	char *inp, *fltr;
	char *newbase, *subbase;
	char buf[1024];
	int e;
	int r;

	if (p_fltr == NULL || *p_fltr == '\0') {
		if (*(inp = ntg_input(ctx, "Search For: ", NULL)) == '\0')
			fltr = NULL;
		else {
			(void)strncpy(fltr_buf, inp, sizeof(fltr_buf) - 1);
			fltr_buf[sizeof(fltr_buf) - 1] = '\0';
			fltr = fltr_buf;
		}
		free(inp);
	} else {
		(void)strncpy(fltr_buf, p_fltr, sizeof(fltr_buf) - 1);
		fltr_buf[sizeof(fltr_buf) - 1] = '\0';
		fltr = fltr_buf;
	}

	e = 0;
	ntg_info_up(ctx, retr_p_msg);
	res = ntl_search(ctx->ld, base, scope, fltr, ctx->cfg->showop ? lda_showop : lda_search);
	ntg_info_down(ctx);
	if (res == NULL) {
		ntg_error_va(ctx, "</U>Error Finding Entries (General Info)<!U>\n%s\n "
		    "* Search Filter: %s\n * Base DN: %s\n * Scope: %s", ntl_error_str,
		    fltr, base, (scope == LDAP_SCOPE_ONELEVEL) ? "ONELEVEL" : "SUBTREE");
		ntg_error_ldap(ctx, "</U>Error Finding Entries (Specific Info)<!U>", "*", ntl_error);
		e = 1;
		goto ntg_search_out;
	} else if (ldap_count_entries(ctx->ld, res) < 1) {
		ntg_message(ctx, nf_msg);
		e = 2;
		goto ntg_search_out;
	}

	attr = ntl_entry(ctx->ld, res);
	(void)ldap_sort_entries(ctx->ld, &res, NULL, (int (*)(const char *, const char *))strcasecmp);
	if (ctx->sortleft) {
		if (ntl_lookup_val(attr, "objectClass", "organization")
		    || ntl_lookup_val(attr, "objectClass", "organizationalUnit")) {
			(void)ldap_sort_entries(ctx->ld, &res, "cn", (int (*)(const char *, const char *))strcasecmp);
			(void)ldap_sort_entries(ctx->ld, &res, "o", (int (*)(const char *, const char *))strcasecmp);
			(void)ldap_sort_entries(ctx->ld, &res, "ou", (int (*)(const char *, const char *))strcasecmp);
		} else {
			/* assume person-Style objectClass */
			(void)ldap_sort_entries(ctx->ld, &res, "cn", (int (*)(const char *, const char *))strcasecmp);
			(void)ldap_sort_entries(ctx->ld, &res, "sn", (int (*)(const char *, const char *))strcasecmp);
		}
	}
	ntl_entry_free(attr);

	(void)snprintf(buf, sizeof(buf), "Base DN: %s, Filter: %s, Scope: %s, Sort: %s%s",
	    base, fltr, (scope == LDAP_SCOPE_ONELEVEL) ? "one" : "sub",
	    ctx->sortleft ? "l" : "r/dn",
	    ctx->cfg->showop ? ", Show Operational Attrs" : "");

	r = (int)ntg_search_select(ctx, ctx->ld, res, base, buf, displayfn);

	if (res != NULL) {
		(void)ldap_msgfree(res);
		res = NULL;
	}

	if (r == NTG_MENU_RET_REFRESH)
		return ntg_search(ctx, base, scope, fltr, displayfn);
	else if (r == NTG_MENU_RET_REFSEARCH) {
		char *newfltr;
		char *refine;

		if (*(refine = ntg_input(ctx, "Refine Search With: ", NULL)) != '\0') {
			if (fltr == NULL) {
				(void)strncpy(fltr_buf, "objectClass=*", sizeof(fltr_buf) - 1);
				fltr_buf[sizeof(fltr_buf) - 1] = '\0';
				fltr = fltr_buf;
			}
			nt_asprintf(&newfltr, "(&%c%s%c%c%s%c)",
			    (*fltr == '(' ? ' ' : '('),
			    fltr,
			    (fltr[strlen(fltr)-1] == ')' ? ' ' : ')'),
			    (*refine == '(' ? ' ' : '('),
			    refine,
			    (refine[strlen(refine)-1] == ')' ? ' ' : ')'));
			free(refine);
			/* start a refined-search with refined filter, recognize any suicide */
			if (ntg_search(ctx, base, scope, newfltr, displayfn) == NTG_MENU_RET_SUICIDE)
				e = NTG_MENU_RET_SUICIDE;
			else {
				free(newfltr);
				return ntg_search(ctx, base, scope, fltr, displayfn);
			}
			free(newfltr);
		} else {
			ntg_message(ctx, nval_msg);
			free(refine);
			return ntg_search(ctx, base, scope, fltr, displayfn);
		}
	} else if (r == NTG_MENU_RET_SUBSEARCH) {
		char *newfltr;

		if (*(newfltr = ntg_input(ctx, "Search Subtree For: ", NULL)) != '\0') {
			/* start a subtree-search with new filter, recognize any suicide */
			if (ntg_search(ctx, base, LDAP_SCOPE_SUBTREE, newfltr, displayfn) == NTG_MENU_RET_SUICIDE)
				e = NTG_MENU_RET_SUICIDE;
			else {
				free(newfltr);
				return ntg_search(ctx, base, scope, fltr, displayfn);
			}
			free(newfltr);
		} else {
			free(newfltr);
			ntg_message(ctx, nval_msg);
			return ntg_search(ctx, base, scope, fltr, displayfn);
		}
	} else if (r == NTG_MENU_RET_CHBASE) {
		if (*(subbase = ntg_input(ctx, "Sub Base: ", NULL)) != '\0') {
			nt_asprintf(&newbase, "%s,%s", subbase, (base == NULL) ? "" : base);
			free(subbase);
			/* start a sub-search with new base, recognize any suicide */
			if (ntg_search(ctx, newbase, scope, fltr, displayfn) == NTG_MENU_RET_SUICIDE) {
				free(newbase);
				e = NTG_MENU_RET_SUICIDE;
			} else {
				free(newbase);
				return ntg_search(ctx, base, scope, fltr, displayfn);
			}
		} else {
			ntg_message(ctx, nval_msg);
			free(subbase);
			return ntg_search(ctx, base, scope, fltr, displayfn);
		}
	} else if (r == NTG_MENU_RET_CHSORT) {
		ctx->sortleft = !ctx->sortleft;
		return ntg_search(ctx, base, scope, fltr, displayfn);
	} else if (r == NTG_MENU_RET_SCOPE) {
		if (scope == LDAP_SCOPE_SUBTREE)
			scope = LDAP_SCOPE_ONELEVEL;
		else
			scope = LDAP_SCOPE_SUBTREE;
		return ntg_search(ctx, base, scope, fltr, displayfn);
	} else if (r == NTG_MENU_RET_DIVE) {
		if (ctx->tmp != NULL) {
			newbase = ctx->tmp;
			ctx->tmp = NULL;
			/* start a sub-search with new base, recognize any suicide */
			if (ntg_search(ctx, newbase, scope, fltr, displayfn) == NTG_MENU_RET_SUICIDE)
				e = NTG_MENU_RET_SUICIDE;
			free(newbase);
		}
		if (e != NTG_MENU_RET_SUICIDE)
			return ntg_search(ctx, base, scope, fltr, displayfn);
	} else if (r == NTG_MENU_RET_SUICIDE)
		e = NTG_MENU_RET_SUICIDE;

ntg_search_out:
	if (res != NULL)
		(void)ldap_msgfree(res);
	return e;
}

int
ntg_search_select(NTG_CTX *ctx, LDAP *ld, LDAPMessage *res, const char *base, const char *searchdesc, int (*displayfn)(NTG_CTX *, const char *, const char *, int))
{
	LDAPMessage *entry;
	MENU *menu;
	ITEM **items;
	ITEM *choice;
	char *dn;
	char **name, **desc;
	NTL_ATTR *cur_attr;
	const char *cur_a[3];
	char *tmp;
	int count;
	int flags;
	int subflag;
	int i, j;
	int r;

	count = ldap_count_entries(ld, res);
	name = nt_calloc((size_t)count + 2, sizeof(char *));
	desc = nt_calloc((size_t)count + 2, sizeof(char *));

	name[0] = nt_strdup("Return");
	desc[0] = nt_strdup("");
	for (
		entry = ldap_first_entry(ld, res), i = 1;
		entry != NULL;
		entry = ldap_next_entry(ld, entry), i++
	) {
		cur_attr = ntl_entry(ld, entry);
		desc[i] = ldap_get_dn(ld, entry);

		subflag = 0;
		if (ntl_lookup_val(cur_attr, "hasSubordinates", "TRUE"))
			subflag = 1;

		if (ntl_lookup_val(cur_attr, "objectClass", "organizationalUnit")) {
			cur_a[0] = *ntl_lookup(cur_attr, "ou");
			cur_a[1] = *ntl_lookup(cur_attr, "o");
			nt_asprintf(&name[i], "%s ou: %-24s  o: %-10s",
			    subflag ? "+" : " ",
			    (cur_a[0] == NULL ? "(none)" : cur_a[0]),
			    (cur_a[1] == NULL ? "(none)" : cur_a[1]));
		} else if (ntl_lookup_val(cur_attr, "objectClass", "organization")) {
			cur_a[0] = *ntl_lookup(cur_attr, "o");
			cur_a[1] = *ntl_lookup(cur_attr, "cn");
			nt_asprintf(&name[i], "%s o:  %-24s cn: %-10s",
			    subflag ? "+" : " ",
			    (cur_a[0] == NULL ? "(none)" : cur_a[0]),
			    (cur_a[1] == NULL ? "(none)" : cur_a[1]));
		} else {
			/* assume person-Style objectClass */
			cur_a[0] = *ntl_lookup(cur_attr, "sn");
			cur_a[1] = *ntl_lookup(cur_attr, "cn");
			nt_asprintf(&name[i], "%s sn: %-12s cn: %-22s",
			    subflag ? "+" : " ",
			    (cur_a[0] == NULL ? "(none)" : cur_a[0]),
			    (cur_a[1] == NULL ? "(none)" : cur_a[1]));
		}
		ntl_entry_free(cur_attr);

		for (j = 0; name[i][j] != '\0'; j++)
			if (name[i][j] < 32 || name[i][j] > 126)
				name[i][j] = '?';
	}
	name[i] = desc[i] = NULL;

	items = ntg_items((const char **)name, (const char **)desc);
	(void)item_opts_off(items[0], O_SELECTABLE);
	menu = ntg_menu(ctx, items);
	(void)menu_opts_off(menu, O_ONEVALUE);

	for (;;) {
		r = 0;
		(void)post_menu(menu);
		ntg_title(ctx, "Search Results");
		ntg_subtitle(ctx, searchdesc);
		ntg_keyhelp(ctx, (const char **)ctx->keys->keyhelp_search);
		menu_opts_on(menu, O_NONCYCLIC);
		choice = ntg_menu_driver(ctx, menu, NTG_MENU_LDAP);

		if (choice == items[0] || (choice == NULL && ctx->tmpreturn == 0)) {
			break;
		} else if (choice == NULL) {
			if (ctx->tmpreturn == NTG_MENU_RET_REFRESH) {
				r = NTG_MENU_RET_REFRESH;
				break;
			} else if (ctx->tmpreturn == NTG_MENU_RET_REFSEARCH) {
				r = NTG_MENU_RET_REFSEARCH;
				break;
			} else if (ctx->tmpreturn == NTG_MENU_RET_SUBSEARCH) {
				r = NTG_MENU_RET_SUBSEARCH;
				break;
			} else if (ctx->tmpreturn == NTG_MENU_RET_CHBASE) {
				r = NTG_MENU_RET_CHBASE;
				break;
			} else if (ctx->tmpreturn == NTG_MENU_RET_CHSORT) {
				r = NTG_MENU_RET_CHSORT;
				break;
			} else if (ctx->tmpreturn == NTG_MENU_RET_SCOPE) {
				r = NTG_MENU_RET_SCOPE;
				break;
			} else if (ctx->tmpreturn == NTG_MENU_RET_DIVE) {
				ctx->tmp = nt_strdup(item_description(current_item(menu)));
				r = NTG_MENU_RET_DIVE;
				break;
			} else if (ctx->tmpreturn == NTG_MENU_RET_LDIF) {
				(void)ntg_display_ldif(ctx, item_description(current_item(menu)));
			} else if (ctx->tmpreturn == NTG_MENU_RET_CREATE_FROM_TPL) {
				(void)ntg_newobject_tpl(ctx, base);
			} else if (ctx->tmpreturn == NTG_MENU_RET_CREATE_FROM_OC) {
				(void)ntg_newobject_oc(ctx, base);
			} else if (ctx->tmpreturn == NTG_MENU_RET_DEL) {
				if (ntg_yesno(ctx, quest_ays))
					(void)ntg_del(ctx, item_description(current_item(menu)));
			} else if (ctx->tmpreturn == NTG_MENU_RET_DELM) {
				if (ntg_yesno(ctx, quest_ays))
					for (i = 1; i < item_count(menu) && items[i] != NULL; i++)
						if (item_value(items[i]) == TRUE)
							(void)ntg_del(ctx, item_description(items[i]));
			} else if (ctx->tmpreturn == NTG_MENU_RET_COPY) {
				if (*(tmp = ntg_input(ctx, inp_newdn, item_description(current_item(menu)))) != '\0')
					(void)ntg_copy(ctx, item_description(current_item(menu)), tmp);
				free(tmp);
			} else if (ctx->tmpreturn == NTG_MENU_RET_COPYM) {
				if (*(tmp = ntg_input(ctx, inp_targetbase, NULL)) != '\0') {
					for (i = 1; i < item_count(menu) && items[i] != NULL; i++)
						if (item_value(items[i]) == TRUE)
							(void)ntg_copy_base(ctx, item_description(items[i]), tmp);
				}
				free(tmp);
			} else if (ctx->tmpreturn == NTG_MENU_RET_MOVE) {
				if (*(tmp = ntg_input(ctx, inp_newloc, item_description(current_item(menu)))) != '\0')
					(void)ntg_move(ctx, item_description(current_item(menu)), tmp);
				free(tmp);
			} else if (ctx->tmpreturn == NTG_MENU_RET_MOVEM) {
				if (*(tmp = ntg_input(ctx, inp_targetbase, NULL)) != '\0') {
					for (i = 1; i < item_count(menu) && items[i] != NULL; i++)
						if (item_value(items[i]) == TRUE)
							(void)ntg_move_base(ctx, item_description(items[i]), tmp);
				}
				free(tmp);
			} else if (ctx->tmpreturn == NTG_MENU_RET_SUICIDE) {
				r = NTG_MENU_RET_SUICIDE;
				break;
			}
		} else if (choice != NULL) {
			flags = 0;

			if (ctx->tmpflags) {
				flags = ctx->tmpflags;
				ctx->tmpflags = 0;
			}

			dn = nt_strdup(item_description(choice));
			while ((r = displayfn(ctx, base, dn, flags)) != 0) {
				if (r == NTG_PAGER_RET_PREV) {
					if ((item_index(current_item(menu)) < 2) || (menu_driver(menu, REQ_PREV_ITEM) != E_OK))
						ntg_message(ctx, nfu_msg);
				} else if (r == NTG_PAGER_RET_NEXT) {
					if (menu_driver(menu, REQ_NEXT_ITEM) != E_OK)
						ntg_message(ctx, nfu_msg);
				} else if (r == NTG_PAGER_RET_RELOAD) {
					;
				} else if (r == NTG_PAGER_RET_SUICIDE) {
					free(dn);
					r = NTG_MENU_RET_SUICIDE;
					goto ntg_search_select_out;
				}
				free(dn);
				dn = nt_strdup(item_description(current_item(menu)));
			}
			free(dn);
		}
		(void)unpost_menu(menu);
	}

ntg_search_select_out:
	(void)unpost_menu(menu);
	(void)free_menu(menu);
	ntg_items_free(items);

	for (i = 0; name[i] != NULL; i++)
		free(name[i]);
	free(name);
	for (i = 0; desc[i] != NULL; i++)
		free(desc[i]);
	free(desc);

	return r;
}

int
ntg_display_ldif(NTG_CTX *ctx, const char *dn)
{
	LDAPMessage *res, *entry;
	NTL_ATTR *attr;
	char title[128];
	char **ldif;

	*title = '\0';
	(void)snprintf(title, sizeof(title), "LDIF: %s", dn);
	ntg_title(ctx, title);
	ntg_keyhelp(ctx, NULL);
	wrefresh(stdscr);

	ntg_info_up(ctx, retr_msg);

	if ((res = ntl_search(ctx->ld, dn, LDAP_SCOPE_BASE, NULL, ctx->cfg->showop ? lda_showop : NULL)) == NULL) {
		ntg_info_down(ctx);
		ntg_message(ctx, nf_msg);
		goto ntg_display_ldif_out;
	}

	if ((entry = ldap_first_entry(ctx->ld, res)) == NULL) {
		ntg_info_down(ctx);
		ntg_message(ctx, nf_msg);
		goto ntg_display_ldif_out;
	}

	attr = ntl_entry(ctx->ld, entry);
	ldif = ntl_ldif_entry(attr);

	ntg_info_down(ctx);

	ntg_display(ctx, (const char **)ldif);

	ntl_entry_free(attr);
	ntl_ldif_free(ldif);
ntg_display_ldif_out:
	(void)ldap_msgfree(res);
	return 0;
}

static FIELD *
f_label(const char *s, int y, int x, int width)
{
	FIELD *f;

	if ((f = new_field(1, width, y, x, 0, 0)) != NULL) {
		(void)set_field_userptr(f, NULL);
		(void)field_opts_off(f, O_ACTIVE | O_EDIT);
		(void)set_field_buffer(f, 0, s);
	}
	return f;
}

static FIELD *
f_field(const char *s, const char *a, int type, int null, int y, int x, int width, int prec, const char *regex, const char **en)
{
	FIELD *f;
	char *p;

	if ((f = new_field(1, width, y, x, 0, 1)) != NULL) {
		(void)set_field_userptr(f, (void *)a);
		if (null)
			(void)field_opts_on(f, O_NULLOK);
		else
			(void)field_opts_off(f, O_NULLOK);

		switch (type) {
		case NTL_T_MAIL:
			(void)set_field_type(f, TYPE_REGEXP, NTL_R_MAIL);
			(void)set_field_buffer(f, 1, "mail");
			break;
		case NTL_T_NAME:
			(void)set_field_type(f, TYPE_REGEXP, NTL_R_NAME);
			(void)set_field_buffer(f, 1, "name");
			break;
		case NTL_T_CN:
			(void)set_field_type(f, TYPE_REGEXP, NTL_R_CN);
			(void)set_field_buffer(f, 1, "cn");
			break;
		case NTL_T_INT:
			(void)set_field_type(f, TYPE_INTEGER, 0, INT_MIN, INT_MAX);
			(void)set_field_buffer(f, 1, "int");
			break;
		case NTL_T_NUM:
			(void)set_field_type(f, TYPE_NUMERIC, prec, DBL_MIN, DBL_MAX);
			(void)set_field_buffer(f, 1, "num");
			break;
		case NTL_T_BOOL:
			(void)set_field_type(f, TYPE_ENUM, boollist, FALSE, TRUE);
			(void)set_field_buffer(f, 1, "bool");
			break;
		case NTL_T_ALPHA:
			(void)set_field_type(f, TYPE_ALPHA, 1);
			(void)set_field_buffer(f, 1, "alpha");
			break;
		case NTL_T_ALNUM:
			(void)set_field_type(f, TYPE_ALNUM, 1);
			(void)set_field_buffer(f, 1, "alnum");
			break;
		case NTL_T_IPV4:
			(void)set_field_type(f, TYPE_IPV4, 1);
			(void)set_field_buffer(f, 1, "ipv4");
			break;
		case NTL_T_REGEX:
			(void)set_field_type(f, TYPE_REGEXP, regex);
			(void)set_field_buffer(f, 1, "regex");
			break;
		case NTL_T_ENUM:
			(void)set_field_type(f, TYPE_ENUM, en, FALSE, TRUE);
			(void)set_field_buffer(f, 1, "enum");
			break;
		default:
			(void)set_field_buffer(f, 1, "");
			break;
		}
		(void)field_opts_off(f, O_STATIC);
		(void)set_max_field(f, 32768);
		if (s != NULL) {
			p = nt_strdup(s);
			ntl_out_nl(p);
			(void)set_field_buffer(f, 0, p);
			free(p);
		} else
			(void)set_field_buffer(f, 0, "");
	}
	return f;
}

int
ntg_pager(NTG_CTX *ctx, FIELD **f)
{
	FORM *form;
	int r;

	form = new_form(f);
	(void)set_form_sub(form, ctx->mainwin);
	(void)post_form(form);
	ctx->oldcursor = curs_set(1);
	r = ntg_form_driver(ctx, form);

	/* workaround (n)curses Bug/Feature */
	(void)form_driver(form, REQ_NEXT_FIELD);
	(void)form_driver(form, REQ_PREV_FIELD);

	(void)unpost_form(form);
	(void)free_form(form);
	(void)curs_set(ctx->oldcursor);
	return r;
}

int
ntg_passwd(NTG_CTX *ctx, const char *dn)
{
	char *newpass;
	char *val[2];
	char salt[20], *sp;
	char *p;
	int e;

	if (*(newpass = ntg_input(ctx, inp_pass, NULL)) == '\0') {
		free(newpass);
		ntg_message(ctx, nval_msg);
		return -1;
	}

	if (strcmp(newpass, "random") == 0) {
		free(newpass);
		newpass = nt_calloc(9, sizeof(char));
		newpass[8] = '\0';
		ntu_to64(newpass, random(), 4);
		ntu_to64(newpass+4, random(), 4);
		ntg_message_va(ctx, "Random Password Is: %s", newpass);
	}

	ntg_info_up(ctx, wr_msg);

	sp = nt_calloc(9, sizeof(char));
	sp[8] = '\0';
	ntu_to64(sp, lrand48(), 4);
	ntu_to64(sp+4, lrand48(), 4);
	(void)snprintf(salt, sizeof(salt), "$1$%s$", sp);
	free(sp);
	if ((p = crypt(newpass, salt)) == NULL) {
		ntg_message(ctx, nval_msg);
		e = -1;
		goto ntg_passwd_out;
	}
	nt_asprintf(&val[0], "{crypt}%s", p);
	val[1] = NULL;

	e = 0;
	if ((e = ntl_mod(ctx->ld, dn, "userPassword", (const char **)val, LDAP_MOD_REPLACE)) != 0)
		ntg_error_ldap(ctx, "Error Writing Password", dn, e);

ntg_passwd_out:
	free(newpass);
	ntg_info_down(ctx);

	return e;
}

int
ntg_modrdn(NTG_CTX *ctx, const char *dn)
{
	char *msg;
	char *newrdn;
	int e;

	e = -1;
	nt_asprintf(&msg, "New RDN for DN \"%s\": ", dn);
	if (*(newrdn = ntg_input(ctx, msg, NULL)) != '\0') {
		ntg_info_up(ctx, mod_msg);
		if ((e = ldap_modrdn_s(ctx->ld, dn, newrdn)) != LDAP_SUCCESS)
			ntg_error_ldap(ctx, "Error Performing Mod RDN", newrdn, e);
		ntg_info_down(ctx);
	}
	free(msg);
	free(newrdn);

	return e;
}

int
ntg_del(NTG_CTX *ctx, const char *dn)
{
	int e;

	ntg_info_up(ctx, dl_msg);
	if ((e = ntl_del(ctx->ld, dn)) != LDAP_SUCCESS)
		ntg_error_ldap(ctx, "Error Deleting Entry", dn, e);
	ntg_info_down(ctx);

	return e;
}

int
ntg_copy_base(NTG_CTX *ctx, const char *source_dn, const char *target_base)
{
	int e;

	ntg_info_up(ctx, cp_msg);
	if ((e = ntl_copy_base(ctx->ld, source_dn, target_base)) != 0)
		ntg_error_ldap(ctx, "Error Copying Entry", target_base, e);
	ntg_info_down(ctx);

	return e;
}

int
ntg_copy(NTG_CTX *ctx, const char *source_dn, const char *dest_dn)
{
	int e;

	ntg_info_up(ctx, cp_msg);
	if ((e = ntl_copy(ctx->ld, source_dn, dest_dn)) != 0)
		ntg_error_ldap(ctx, "Error Copying Entry", dest_dn, e);
	ntg_info_down(ctx);

	return e;
}

int
ntg_move_base(NTG_CTX *ctx, const char *source_dn, const char *target_base)
{
	int e;

	ntg_info_up(ctx, mv_msg);
	if ((e = ntl_move_base(ctx->ld, source_dn, target_base)) != 0)
		ntg_error_ldap(ctx, "Error Moving Entry", target_base, e);
	ntg_info_down(ctx);

	return e;
}

int
ntg_move(NTG_CTX *ctx, const char *source_dn, const char *dest_dn)
{
	int e;

	ntg_info_up(ctx, mv_msg);
	if ((e = ntl_move(ctx->ld, source_dn, dest_dn)) != 0)
		ntg_error_ldap(ctx, "Error Moving Entry", dest_dn, e);
	ntg_info_down(ctx);

	return e;
}

int
ntg_pager_write(NTG_CTX *ctx, const char *dn, FIELD **f)
{
	NTL_ATTR *chgset, *mdel;
	char *attr;
	char *p;
	int i, e, errors;

	errors = 0;

	ntg_info_up(ctx, wr_msg);

	chgset = nt_calloc(1024, sizeof(NTL_ATTR));
	mdel = nt_calloc(1024, sizeof(NTL_ATTR));
	chgset[0].name = mdel[0].name = NULL;
	chgset[0].val = mdel[0].val = NULL;
	chgset[0].tsize = mdel[0].tsize = 1024;

	/*
	 * Step 1:
	 * Get list of fields that have been changed and add them to chgset.
	 */
	for (i = 0; f[i] != NULL; i++) {
		if (field_status(f[i]) == TRUE) {
			attr = field_userptr(f[i]);
			p = ntl_strip(field_buffer(f[i], 0));
			ntl_in_nl(p);
			if (*p == '\0') {
				/*
				 * We have an empty field:
				 * We don't add ourself to the structure but record
				 * that the attribute value has been erased.
				 */
				ntl_attr_add(mdel, attr, "empty");
			} else
				ntl_attr_add(chgset, attr, p);
			free(p);
		}
	}
	/*
	 * Step 2:
	 * Get related fields that are possibly not marked as changed but need to be re-committed.
	 * Add them to chgset, too.  (But only if != the empty string.)
	 */
	for (i = 0; f[i] != NULL; i++) {
		attr = field_userptr(f[i]);
		p = ntl_strip(field_buffer(f[i], 0));
		ntl_in_nl(p);
		if (*p && attr != NULL) {
			if (*ntl_lookup(chgset, attr) != NULL)
				ntl_attr_add(chgset, attr, p);
			if (*ntl_lookup(mdel, attr) != NULL)
				ntl_attr_add(chgset, attr, p);
		}
		free(p);
	}
	/*
	 * Step 3:
	 * Propagate each modification change.
	 */
	for (i = 0; chgset[i].name != NULL; i++)
		if ((e = ntl_mod(ctx->ld, dn, chgset[i].name, (const char **)chgset[i].val, LDAP_MOD_REPLACE)) != 0) {
			ntg_error_ldap(ctx, "Error Propagating Changes, Step 3: Modifications", dn, e);
			errors++;
		}
	/*
	 * Step 4:
	 * Determine and propagate needed attribute deletions.
	 */
	for (i = 0; mdel[i].name != NULL; i++)
		if (*ntl_lookup(chgset, mdel[i].name) == NULL) {
			/*
			 * no attribute values left, delete
			 */
			if ((e = ntl_mod(ctx->ld, dn, mdel[i].name, NULL, LDAP_MOD_DELETE)) != 0) {
				ntg_error_ldap(ctx, "Error Propagating Changes, Step 4: Deletions", dn, e);
				errors++;
			}
		}

	ntl_entry_free(chgset);
	ntl_entry_free(mdel);

	ntg_info_down(ctx);

	return errors;
}

int
ntg_create_obj_from_tpl(NTG_CTX *ctx, const char *base, const NTL_TPLC *tplc)
{
	NTL_ATTR *new;
	char *key = NULL;
	char *inp, *tmp, *tmp3;
	const char **tmp2;
	char buf[256];
	char *p;
	int i, j;
	int e = 0;

	if (tplc == NULL) {
		ntg_message(ctx, ntplc_msg);
		return -1;
	}

	/* copy fix part of template in new */
	new = nt_calloc((size_t)tplc->crmax + tplc->usrmax + 1, sizeof(NTL_ATTR));
	new[0].name = NULL;
	new[0].val = NULL;
	new[0].tsize = tplc->crmax + tplc->usrmax + 1;
	for (i = 0; tplc->crattr[i].name != NULL; i++)
		for (j = 0; tplc->crattr[i].val != NULL && tplc->crattr[i].val[j] != NULL; j++)
			ntl_attr_add(new, tplc->crattr[i].name, tplc->crattr[i].val[j]);

	/* add rest to new */
	for (i = 0; tplc->usrattr[i].name != NULL ; i++) {
		if (strcmp("?", tplc->usrattr[i].val[0]) == 0) {
			/* ask for input */
			ntg_attribute_tip(ctx, tplc->usrattr[i].name);
			nt_asprintf(&tmp, "Enter Value For \"%s\": ", tplc->usrattr[i].name);
			inp = ntg_input(ctx, tmp, NULL);
			ntg_tip_down(ctx);
			free(tmp);
			if (*inp != '\0') {
				if (i == 0)
					key = nt_strdup(inp);
				ntl_attr_add(new, tplc->usrattr[i].name, inp);
			} else {
				free(inp);
				ntg_message(ctx, nval_msg);
				e = 1;
				break;
			}
			free(inp);
		} else {
			/* append values from other fields as specified */
			inp = nt_strdup("");
			for (j = 0; tplc->usrattr[i].val[j] != NULL; j++) {
				if (*tplc->usrattr[i].val[j] == '"')
					nt_asprintf(&tmp, "%s%s", inp, tplc->usrattr[i].val[j] + 1);
				else if (strcmp("!binddn", tplc->usrattr[i].val[j]) == 0)
					nt_asprintf(&tmp, "%s%s", inp, ctx->cfg->dn);
				else if (strcmp("!unixtime", tplc->usrattr[i].val[j]) == 0)
					nt_asprintf(&tmp, "%s%lu", inp, time(NULL));
				else if (strcmp("!gentime", tplc->usrattr[i].val[j]) == 0) {
					tmp3 = ntu_gentime(time(NULL));
					if (tmp3 != NULL) {
						nt_asprintf(&tmp, "%s%s", inp, tmp3);
						free(tmp3);
					} else
						tmp = nt_strdup(inp);
				} else if (strcmp("!random4", tplc->usrattr[i].val[j]) == 0) {
					buf[4] = '\0';
					ntu_to64(buf, random(), 4);
					nt_asprintf(&tmp, "%s%s", inp, buf);
				} else if (strcmp("!random8", tplc->usrattr[i].val[j]) == 0) {
					buf[8] = '\0';
					ntu_to64(buf, random(), 4);
					ntu_to64(buf+4, random(), 4);
					nt_asprintf(&tmp, "%s%s", inp, buf);
				} else if (strcmp("!random16", tplc->usrattr[i].val[j]) == 0) {
					buf[16] = '\0';
					ntu_to64(buf, random(), 4);
					ntu_to64(buf+4, random(), 4);
					ntu_to64(buf+8, random(), 4);
					ntu_to64(buf+12, random(), 4);
					nt_asprintf(&tmp, "%s%s", inp, buf);
				} else if (strcmp("!hostname", tplc->usrattr[i].val[j]) == 0) {
					if (gethostname(buf, sizeof(buf)) != 0) {
						ntg_error_va(ctx, "</U>Unable To Get Hostname<!U>\n"
						    " * Requested In Template: %s\n * Attribute: %s",
						    tplc->name, tplc->usrattr[i].name);
						tmp = nt_strdup(inp);
					} else
						nt_asprintf(&tmp, "%s%s", inp, buf);
				} else if (strcmp("!domainname", tplc->usrattr[i].val[j]) == 0) {
					if (getdomainname(buf, sizeof(buf)) != 0) {
						ntg_error_va(ctx, "</U>Unable To Get Domainname<!U>\n"
						    " * Requested In Template: %s\n * Attribute: %s",
						    tplc->name, tplc->usrattr[i].name);
						tmp = nt_strdup(inp);
					} else
						nt_asprintf(&tmp, "%s%s", inp, buf);
				} else {
					if (*(tmp2 = ntl_lookup(new, tplc->usrattr[i].val[j])) == NULL) {
						ntg_error_va(ctx, "</U>Error In Template, "
						    "Suggested Append Impossible<!U>\n * Template: %s\n"
						    " * Attribute: %s\n * False Append: %s",
						    tplc->name, tplc->usrattr[i].name, tplc->usrattr[i].val[j]);
						tmp = nt_strdup(inp);
					} else
						nt_asprintf(&tmp, "%s%s", inp, tmp2[0]);
				}
				free(inp);
				inp = tmp;
			}
			ntl_attr_add(new, tplc->usrattr[i].name, inp);
			free(inp);
		}
	}

	if (key == NULL) {
		ntg_message(ctx, nval_msg);
		e = 2;
	}

	if (e == 0) {
		ntg_info_up(ctx, cr_msg);
		nt_asprintf(&p, "%s=%s,%s", tplc->usrattr[0].name, key, base);
		if ((e = ntl_add(ctx->ld, p, new)) != 0)
			ntg_error_ldap(ctx, "Creation Failed", p, e);
		free(p);
		ntg_info_down(ctx);
	}

	ntl_entry_free(new);
	if (key != NULL)
		free(key);

	return e;
}

int
ntg_create_attr(NTG_CTX *ctx, const char *dn, const NTL_TPLC *tplc)
{
	char *attribute;
	char *inp;
	int e;

	if (tplc != NULL) {
		/*
		 * Case 1:
		 * We've got a template for this entry.  Use it.
		 */
		char **attrnames;
		int choice;
		int i, k;

		attrnames = nt_calloc(1, sizeof(char *));
		attrnames[0] = NULL;
		for (i = 0, k = 0; tplc->tpl[i] != NULL; i++)
			if (tplc->tpl[i]->attr != NULL
			    && strlen(tplc->tpl[i]->attr) > 1
			    && strcmp(tplc->tpl[i]->attr, "dn") != 0) {
				attrnames = nt_realloc(attrnames, (k + 2) * sizeof(char *));
				attrnames[k] = tplc->tpl[i]->attr;
				attrnames[k+1] = NULL;
				k++;
			}

		if ((choice = ntg_select(ctx, "Choose An Attribute To Add: ", (const char **)attrnames)) < 0 || choice > k) {
			free(attrnames);
			return -1;
		}

		attribute = nt_strdup(attrnames[choice]);
		free(attrnames);

	} else if (ctx->schema != NULL) {
		/*
		 * Case 2:
		 * We ain't got no template but we could read schema from server.  Yiiiiieeeeha!
		 *
		 * We interpolate the objectClasses of the current entry and determine all possible attributes.
		 * Then, we politely ask the user to choose an AttributeType he wishes to add.
		 */
		LDAPMessage *res, *entry;
		NTL_ATTR *attr;
		const char *s_attr[] = { "objectClass", NULL };
		char **i_ocs, **attrnames;
		int count, choice;
		int try3;

		attribute = NULL;
		res = entry = NULL;
		attr = NULL;
		try3 = 0;

		if ((res = ntl_search(ctx->ld, dn, LDAP_SCOPE_BASE, NULL, s_attr)) == NULL
		    || (entry = ldap_first_entry(ctx->ld, res)) == NULL
		    || (attr = ntl_entry(ctx->ld, entry)) == NULL)
			ntg_error_va(ctx, "Entry Not Found (Disappeared?).");
		else {
			if ((i_ocs = ntls_interpolate_ocs((const NTLS_SCHEMA *)ctx->schema, (const char **)ntl_lookup(attr, "objectClass"))) == NULL) {
				ntg_error_va(ctx, "Unable To Interpolate OCs.");
				try3 = 1;
				goto ntg_create_attr_s2_out;
			}

			if ((attrnames = ntls_acc_ats((const NTLS_SCHEMA *)ctx->schema, (const char **)i_ocs, '*')) == NULL) {
				ntls_ocs_free(i_ocs);
				ntg_error_va(ctx, "Unable To Accumulate ACSs.");
				try3 = 1;
				goto ntg_create_attr_s2_out;
			}

			count = 0;
			while (attrnames[count] != NULL)
				count++;
			qsort(attrnames, (size_t)count, sizeof(char *), ntu_strcasecmp);

			if ((choice = ntg_select(ctx, "Choose An Attribute To Add: ", (const char **)attrnames)) < 0) {
				ntl_entry_free(attr);
				ntls_ocs_free(i_ocs);
				ntls_acc_free(attrnames);
				(void)ldap_msgfree(res);
				return -1;
			}
			attribute = nt_strdup(attrnames[choice]);
			ntls_acc_free(attrnames);
			ntls_ocs_free(i_ocs);
		}

ntg_create_attr_s2_out:
		if (res != NULL)
			(void)ldap_msgfree(res);
		if (attr != NULL)
			ntl_entry_free(attr);

		if (try3)
			goto ntg_create_attr_s3;
		if (attribute == NULL)
			return -1;

	} else {
ntg_create_attr_s3:
		/*
		 * Case 3:
		 * Well, we're sorta out of luck here.
		 * Just ask the user to supply us something.  Anything.
		 */
		if (*(inp = ntg_input(ctx, "Enter Attribute: ", NULL)) == '\0') {
			ntg_message(ctx, nval_msg);
			free(inp);
			return -1;
		}
		attribute = inp;
	}

	e = 0;

	ntg_attribute_tip(ctx, attribute);
	if (*(inp = ntg_input(ctx, "Enter Value: ", NULL)) == '\0') {
		ntg_message(ctx, nval_msg);
		e = -1;
	} else {
		const char *val[2];

		ntg_info_up(ctx, cr_msg);
		val[0] = inp;
		val[1] = NULL;
		if ((e = ntl_mod(ctx->ld, dn, attribute, val, LDAP_MOD_ADD)) != 0) {
			ntg_error_ldap(ctx, "Error Creating Attribute", dn, e);
			e = 1;
		}
		ntg_info_down(ctx);
	}
	ntg_tip_down(ctx);

	free(attribute);
	free(inp);
	return e;
}

const NTL_TPLC *
ntg_get_tplc(NTG_CTX *ctx, const char **oc)
{
	int i, j;

	for (j = 0; ctx->cfg->tplc[j] != NULL && ctx->cfg->tplc[j]->name != NULL; j++)
		for (i = 0; oc[i] != NULL; i++)
			if (strcmp(ctx->cfg->tplc[j]->name, oc[i]) == 0)
				return ctx->cfg->tplc[j];
	return NULL;
}

int
ntg_pager_generic(NTG_CTX *ctx, const char *base, const char *dn, int flags)
{
	LDAPMessage *res, *entry;
	NTL_ATTR *attr;
	const NTL_TPLC *tplc;
	FIELD **f;
	char t[128];
	char *tmp;
	int desc_len, val_len;
	int r;

	ntg_info_up(ctx, retr_msg);
	if ((res = ntl_search(ctx->ld, dn, LDAP_SCOPE_BASE, NULL, ctx->cfg->showop ? lda_showop : NULL)) == NULL) {
		ntg_message(ctx, nf_msg);
		return 0;
	}

	if ((entry = ldap_first_entry(ctx->ld, res)) == NULL) {
		(void)ldap_msgfree(res);
		ntg_info_down(ctx);
		ntg_message(ctx, nf_msg);
		return 0;
	}

	desc_len = 34;
	val_len = (ctx->mainwin_maxx - desc_len < 7) ? 7 : ctx->mainwin_maxx - desc_len;

	if (val_len > 80)
		val_len = 80;

	attr = ntl_entry(ctx->ld, entry);
	(void)ldap_msgfree(res);
	if (attr == NULL) {
		ntg_message(ctx, ie_msg);
		return 0;
	}

	tplc = NULL;
	if (flags == 0) {
		if ((tplc = ntg_get_tplc(ctx, (const char **)ntl_lookup(attr, "objectClass"))) == NULL) {
			ntl_entry_free(attr);
			ntg_message(ctx, ntplf_msg);
			return 0;
		}
		f = ntg_genfields(ctx, attr, dn, (const NTL_TPL **)tplc->tpl, desc_len, val_len, ctx->mainwin_maxy);
		(void)snprintf(t, sizeof(t), "%s: %s", tplc->name, dn);
	} else {
		f = ntg_genfields_auto(ctx, attr, dn, desc_len, val_len, ctx->mainwin_maxy);
		(void)snprintf(t, sizeof(t), "(auto): %s", dn);
	}

	ntg_title(ctx, t);
	ntg_keyhelp(ctx, (const char **)ctx->keys->keyhelp_form);
	ntg_refresh(ctx);

	ntg_info_down(ctx);

	switch (r = ntg_pager(ctx, f)) {
	case NTG_PAGER_RET_WRITE:
		(void)ntg_pager_write(ctx, dn, f);
		break;
	case NTG_PAGER_RET_DEL:
		if (ntg_yesno(ctx, quest_ays))
			(void)ntg_del(ctx, dn);
		break;
	case NTG_PAGER_RET_CREATE_FROM_TPL:
		(void)ntg_create_obj_from_tpl(ctx, base, tplc);
		r = 0;
		break;
	case NTG_PAGER_RET_CREATE_FROM_OC:
		(void)ntg_newobject_oc(ctx, base);
		r = 0;
		break;
	case NTG_PAGER_RET_CREATE_ATTR:
		(void)ntg_create_attr(ctx, dn, tplc);
		break;
	case NTG_PAGER_RET_LDIF:
		(void)ntg_display_ldif(ctx, dn);
		break;
	case NTG_PAGER_RET_SEARCH:
		if (*(tmp = ntg_input(ctx, "Search For (In Current): ", NULL)) != '\0') {
			r = ntg_search(ctx, dn, LDAP_SCOPE_ONELEVEL, tmp, ntg_pager_generic);
			free(tmp);
			if (r == NTG_MENU_RET_SUICIDE)
				r = NTG_PAGER_RET_SUICIDE;
			else if (r != 0) {
				ntg_fields_free(f);
				ntl_entry_free(attr);
				return ntg_pager_generic(ctx, base, dn, 0);
			}
		} else
			free(tmp);
		break;
	case NTG_PAGER_RET_PASSWD:
		(void)ntg_passwd(ctx, dn);
		break;
	case NTG_PAGER_RET_MODRDN:
		(void)ntg_modrdn(ctx, dn);
		r = 0;
		break;
	case NTG_PAGER_RET_COPY:
		if (*(tmp = ntg_input(ctx, inp_newdn, dn)) != '\0')
			(void)ntg_copy(ctx, dn, tmp);
		free(tmp);
		break;
	case NTG_PAGER_RET_MOVE:
		if (*(tmp = ntg_input(ctx, inp_newloc, dn)) != '\0')
			(void)ntg_move(ctx, dn, tmp);
		free(tmp);
		r = 0;
		break;
	case NTG_PAGER_RET_SUICIDE:
		r = NTG_PAGER_RET_SUICIDE;
		break;
	default:
		break;
	}

	/*
	 * ntg_pager_write needs those
	 * (in case we had a ntg_genfields_auto)
	 */
	ntl_entry_free(attr);
	ntg_fields_free(f);

	(void)wrefresh(ctx->mainwin);
	return r;
}

void
ntg_keyhelp(NTG_CTX *ctx, const char **keyhelp)
{
	int width, i;

	ctx->keyhelp = keyhelp;
	width = getmaxx(ctx->bottomwin);

	(void)wattrset(ctx->bottomwin, A_NORMAL | C_EM);
	(void)mvwhline(ctx->bottomwin, 0, 0, 0, width);

	for (i = 0; i < NTG_KEYHELP_LINES; i++)
		(void)mvwhline(ctx->bottomwin, i + 1, 0, ' ', width);

	if (keyhelp != NULL) {
		for (i = 0; i < NTG_KEYHELP_LINES && keyhelp[i] != NULL; i++)
			(void)mvwaddstr(ctx->bottomwin, i + 1, 0, keyhelp[i]);
	}
}

static chtype
ntg_genfields_mapcolor(int c, int invert)
{
	chtype a;

	switch (c) {
	case 1:
		a = (invert ? C_BLUE : C_NORMAL);
		break;
	case 2:
		a = C_TEXT;
		break;
	case 3:
		a = C_STRONG;
		break;
	case 4:
		a = C_EM;
		break;
	default:
		a = C_NORMAL;
	}
	return a;
}

FIELD **
ntg_genfields(NTG_CTX *ctx, const NTL_ATTR *attr, const char *dn, const NTL_TPL **tpl, int desc_len, int val_len, int height)
{
	FIELD **f;
	int i, j, line;
	int flag;
	int pflag;
	const char **val;

	(void)ctx; /* reserved for later use */

	f = nt_calloc(NTG_MAXFIELDS, sizeof(FIELD *));

	i = j = line = 0;
	while (tpl[i] != NULL && j < NTG_MAXFIELDS - 3) {
		if (strcmp(tpl[i]->desc, "") == 0) {
			i++;
			line++;
			continue;
		}

		pflag = 0;
		if (line > height - NTG_GENFIELDS_BOTTOM_SPACE) {
			line = 0;
			pflag = 1;
		}

		f[j++] = f_label(tpl[i]->desc, line, 0, desc_len);
		if (pflag)
			set_new_page(f[j-1], 1);
		if (tpl[i]->a)
			set_field_back(f[j-1], ntg_genfields_mapcolor(tpl[i]->a, 0));

		if (tpl[i]->type != -1) {
			val = ntl_lookup(attr, tpl[i]->attr);
			if (tpl[i]->multi) {
				flag = 0;
				while (*val != NULL && j < NTG_MAXFIELDS - 3) {
					flag = 1;
					if (line > height - NTG_GENFIELDS_BOTTOM_SPACE) {
						line = 0;
						set_new_page(f[j-1], 1);
					}
					f[j++] = f_field(*val++, tpl[i]->attr, tpl[i]->type, tpl[i]->null,
					    line++, desc_len, val_len, tpl[i]->prec, tpl[i]->regex, (const char **)tpl[i]->en);
					if (tpl[i]->a)
						set_field_back(f[j-1], ntg_genfields_mapcolor(tpl[i]->a, 1));
				}
				if (!flag)
					line++;
				f[j++] = f_field(*val++, tpl[i]->attr, tpl[i]->type, tpl[i]->null,
				    line++, desc_len, val_len, tpl[i]->prec, tpl[i]->regex, (const char **)tpl[i]->en);
			} else
				f[j++] = f_field(*val, tpl[i]->attr, tpl[i]->type, tpl[i]->null,
				    line++, desc_len, val_len, tpl[i]->prec, tpl[i]->regex, (const char **)tpl[i]->en);
		} else if (strcmp(tpl[i]->desc, "dn") == 0)
			f[j++] = f_label(dn, line++, desc_len, val_len);
		else if (tpl[i]->attr != NULL) {
			val = ntl_lookup(attr, tpl[i]->attr);
			flag = 0;
			while (*val != NULL && j < NTG_MAXFIELDS - 3) {
				flag = 1;
				if (line > height - NTG_GENFIELDS_BOTTOM_SPACE) {
						line = 0;
						set_new_page(f[j-1], 1);
				}
				f[j++] = f_label(*val++, line++, desc_len, val_len);
				if (tpl[i]->a)
					set_field_back(f[j-1], ntg_genfields_mapcolor(tpl[i]->a, 1));
			}
			if (!flag)
				line++;
		} else
			line++;

		if (tpl[i]->a)
			set_field_back(f[j-1], ntg_genfields_mapcolor(tpl[i]->a, 1));

		i++;
	}
	f[j] = NULL;

	return f;
}

FIELD **
ntg_genfields_auto(NTG_CTX *ctx, const NTL_ATTR *attr, const char *dn, int desc_len, int val_len, int height)
{
	FIELD **f; 
	int i, j, k, line;
	int pflag;

	f = nt_calloc(NTG_MAXFIELDS, sizeof(FIELD *));

	i = j = line = 0;
	while (attr[i].name != NULL && j < NTG_MAXFIELDS - 3) {
		for (k = 0; attr[i].val[k] != NULL; k++) {
			pflag = 0;
			if (line > height - NTG_GENFIELDS_BOTTOM_SPACE) {
				line = 0;
				pflag = 1;
			}

			if (strcmp(attr[i].name, "dn") == 0) {
				f[j++] = f_label(attr[i].name, line, 0, desc_len);
				set_field_back(f[j-1], C_STRONG);
				if (pflag)
					set_new_page(f[j-1], 1);
				f[j++] = f_label(dn, line++, desc_len, val_len);
				set_field_back(f[j-1], C_STRONG);
			} else {
				f[j++] = f_label(attr[i].name, line, 0, desc_len);
				if (pflag)
					set_new_page(f[j-1], 1);

				if (ctx->schema != NULL
				   && ntls_at_human(ntls_get_at_from_name((const LDAPAttributeType **) ctx->schema->ats, attr[i].name)) != 1)
					f[j++] = f_label("(non human-readable value or unknown Syntax OID)", line++, desc_len, val_len);
				else
					f[j++] = f_field(attr[i].val[k], attr[i].name, 0, 1, line++, desc_len, val_len, 0, NULL, NULL);
			}
		}
		i++;
	}
	f[j] = NULL;

	return f;
}

void
ntg_fields_free(FIELD **fields)
{
	int i;

	if (fields == NULL)
		return;

	for (i = 0; fields[i] != NULL; i++)
		(void)free_field(fields[i]);
	free(fields);
}
