/*
 * $Id: beb5e1c0527e135b1914c5699f943ca8b4e577c1 $
 *
 * This file is part of ntldap.
 *
 * Copyright (c) 2006, 2007, 2008 Niklas Olmes <niklas@noxa.de>
 *                                http://www.nevm.de/ntldap/
 *                                NEVM-RIPE
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <err.h>

#include "ldap.h"

#include "ntl.h"
#include "ntlc.h"
#include "nt_alloc.h"

#define NTL_MAX_ATTRS	1024
#define NTL_MAX_VALS	512
#define NTL_MAX_MODS	1024

static void ntl_attr_add_generic(NTL_ATTR *, const char *, const char *, int);
static const char **ntl_lookup_fn(const NTL_ATTR *, const char *, int (*)(const char *, const char *));
static int ntl_lookup_val_fn(const NTL_ATTR *, const char *, const char *, int (*)(const char *, const char *));

static const char rcsid[] = "$Id: beb5e1c0527e135b1914c5699f943ca8b4e577c1 $";

static const char *empty[] = { NULL };

static char str_ldap[] = "ldap";

int ntl_error = 0;
const char *ntl_error_str = NULL;

LDAP *
ntl_connect(const NTL_CFG *cfg)
{
	int e;
	LDAP *ld;
	int protocol;

#ifdef LDAP_OLD_STYLE
	if ((ld = ldap_init(cfg->host, cfg->port == 0 ? LDAP_PORT : cfg->port)) == NULL) {
		ntl_error = 1;
		ntl_error_str = "Cannot intitialize.\n";
		return NULL;
	}
#else
	LDAPURLDesc url;
	char *ldapuri;

	/* construct URI */
	(void)memset(&url, 0, sizeof(url));
	url.lud_scheme = str_ldap;
	url.lud_host = cfg->host;
	url.lud_port = cfg->port == 0 ? LDAP_PORT : cfg->port;
	url.lud_scope = LDAP_SCOPE_DEFAULT;
	ldapuri = ldap_url_desc2str(&url);

	ld = NULL;
	if ((e = ldap_initialize(&ld, ldapuri)) != LDAP_SUCCESS) {
		ntl_error = e;
		ntl_error_str = ldap_err2string(e);
		free(ldapuri);
		return NULL;
	}
	free(ldapuri);
#endif

	protocol = LDAP_VERSION3;
	if ((e = ldap_set_option(ld, LDAP_OPT_PROTOCOL_VERSION, &protocol)) != LDAP_OPT_SUCCESS) {
		ntl_error = e;
		ntl_error_str = ldap_err2string(e);
		return NULL;
	}

	if (cfg->secure && (e = ldap_start_tls_s(ld, NULL, NULL)) != LDAP_SUCCESS) {
		ntl_error = e;
		ntl_error_str = ldap_err2string(e);
		return NULL;
	}

	if ((e = ldap_simple_bind_s(ld, cfg->dn, cfg->cred)) != LDAP_SUCCESS) {
		ntl_error = e;
		ntl_error_str = ldap_err2string(e);
		return NULL;
	}

	return ld;
}

LDAPMessage *
ntl_search(LDAP *ld, const char *base, int scope, const char *fltr, const char **attrs)
{
	LDAPMessage *res;
	int e;

	if ((e = ldap_search_s(ld, base, scope, fltr, (char **)attrs, 0, &res)) != LDAP_SUCCESS) {
		ntl_error = e;
		ntl_error_str = ldap_err2string(e);
		res = NULL;
	}

	return res;
}

char **
ntl_ldif_entry(const NTL_ATTR *attr)
{
	int i, j, k;
	char **ldif_a;
	char **ldif_e;

	ldif_e = nt_calloc(1, sizeof(char *));
	for (k = 0, i = 0; attr[i].name != NULL; i++) {
		ldif_a = ntl_ldif_attr(attr[i]);
		for (j = 0; ldif_a[j] != NULL; j++) {
			ldif_e = nt_realloc(ldif_e, sizeof(char *)* (k + 2));
			ldif_e[k] = nt_strdup(ldif_a[j]);
			k++;
		}
		ntl_ldif_free(ldif_a);
	}
	ldif_e[k] = NULL;

	return ldif_e;
}

char **
ntl_ldif_attr(const NTL_ATTR attr)
{
	char **ldif;
	int i;

	ldif = nt_calloc(1, sizeof(char *));
	for (i = 0; attr.val[i] != NULL || i == 0; i++) {
		ldif = nt_realloc(ldif, sizeof(char *)* (i + 2));
		nt_asprintf(&ldif[i], "%s: %s", attr.name, attr.val[i]);
	}
	ldif[i] = NULL;

	return ldif;
}

void
ntl_ldif_free(char **ldif)
{
	int i;

	if (ldif == NULL)
		return;

	for (i = 0; ldif[i] != NULL; i++)
		free(ldif[i]);
	free(ldif);
	ldif = NULL;
}

NTL_ATTR *
ntl_entry(LDAP *ld, LDAPMessage *entry)
{
	BerElement *ber;
	char *attr;
	NTL_ATTR *r;
	int i;

	r = nt_calloc(NTL_MAX_ATTRS, sizeof(NTL_ATTR));
	r[0].tsize = NTL_MAX_ATTRS;

	r[0].name = nt_strdup("dn");
	r[0].val = nt_calloc(2, sizeof(char *));
	r[0].vsize = 2;
	/* XXX:
	 * r[0].val[0] = nt_strdup(ldap_get_dn(ld, entry)); // see ldap_get_dn(3), free'd on ldap_memfree(3), but we free ourselves
	 *
	 * But it doesn't free it.  We leak with above.
	 */
	r[0].val[0] = ldap_get_dn(ld, entry);
	r[0].val[1] = NULL;
	for (
		attr = ldap_first_attribute(ld, entry, &ber), i = 1;
		attr != NULL && i < NTL_MAX_ATTRS - 1;
		attr = ldap_next_attribute(ld, entry, ber), i++
	) {
		r[i].name = attr;
		r[i].vsize = NTL_MAX_VALS;
		r[i].val = ldap_get_values(ld, entry, attr);
		r[i].val = nt_realloc(r[i].val, NTL_MAX_VALS * sizeof(char *));
	}
	r[i].name = NULL;
	r[i].val = NULL;

	ber_free(ber, 0);

	return r;
}

void
ntl_entry_free(NTL_ATTR *e)
{
	int i, j;

	if (e == NULL)
		return;

	for (i = 0; e[i].name != NULL; i++) {
		free(e[i].name);
		for (j = 0; e[i].val[j] != NULL; j++)
			free(e[i].val[j]);
		free(e[i].val);
	}
	free(e);
	e = NULL;
}

static const char **
ntl_lookup_fn(const NTL_ATTR *attr, const char *wanted, int fn(const char *, const char *))
{
	char **r;
	int i;

	if (attr == NULL)
		return empty;

	i = 0;
	while (attr[i].name != NULL && fn(attr[i].name, wanted) != 0)
		i++;

	if (attr[i].name != NULL) {
		/* found */
		r = attr[i].val;
	} else
		return empty;

	return (const char **)r;
}

const char **
ntl_lookup(const NTL_ATTR *attr, const char *wanted)
{
	return ntl_lookup_fn(attr, wanted, strcasecmp);
}

static int
ntl_lookup_val_fn(const NTL_ATTR *attr, const char *wanted, const char *wanted_val, int fn(const char *, const char *))
{
	const char **v;

	v = ntl_lookup_fn(attr, wanted, fn);
	while (*v != NULL && strcasecmp(*v, wanted_val) != 0)
		v++;
	if (*v == NULL)
		return 0;
	return 1; /* found */
}

int
ntl_lookup_val(const NTL_ATTR *attr, const char *wanted, const char *wanted_val)
{
	return ntl_lookup_val_fn(attr, wanted, wanted_val, strcasecmp);
}

int
ntl_mod(LDAP *ld, const char *dn, const char *attr, const char **val, int op)
{
	LDAPMod **ldmod;
	char **p;
	int size, i;
	int r;

	ldmod = nt_calloc(2, sizeof(LDAPMod *));

	if (val != NULL) {
		size = 0;
		while (val[size] != NULL)
			size++;
		p = nt_calloc((size_t)size + 1, sizeof(char *));
		for (i = 0; i < size; i++)
			p[i] = nt_strdup(val[i]);
		val[size] = NULL;
	} else
		p = NULL;

	ldmod[0] = nt_calloc(1, sizeof(LDAPMod));
	ldmod[0]->mod_op = op;
	ldmod[0]->mod_type = (attr == NULL) ? NULL : nt_strdup(attr);
	ldmod[0]->mod_values = p;
	ldmod[1] = NULL;

	r = ldap_modify_s(ld, dn, ldmod);
	ldap_mods_free(ldmod, 1);

	if (r == LDAP_SUCCESS)
		return 0;
	return r;
}

int
ntl_add(LDAP *ld, const char *dn, const NTL_ATTR *attr)
{
	LDAPMod *ldmod[NTL_MAX_MODS];
	int r;
	int i;

	for (i = 0; i < NTL_MAX_MODS - 1 && attr[i].name != NULL; i++) {
		ldmod[i] = nt_calloc(1, sizeof(LDAPMod));
		ldmod[i]->mod_op = LDAP_MOD_ADD;
		ldmod[i]->mod_type = attr[i].name;
		ldmod[i]->mod_values = attr[i].val;
	}
	ldmod[i] = NULL;

	r = ldap_add_s(ld, dn, ldmod);

	for (i--; i >= 0; i--)
		free(ldmod[i]);

	if (r == LDAP_SUCCESS)
		return 0;
	return r;
}

int
ntl_copy_base(LDAP *ld, const char *source_dn, const char *dest_base)
{
	LDAPMessage *res, *entry;
	NTL_ATTR *attr;
	char **dnparts;
	char *dest_dn;
	int e;

	if ((res = ntl_search(ld, source_dn, LDAP_SCOPE_BASE, NULL, NULL)) == NULL) {
		ntl_error = -1;
		ntl_error_str = "Entry not found.";
		return ntl_error;
	} else if ((entry = ldap_first_entry(ld, res)) == NULL) {
		(void)ldap_msgfree(res);
		ntl_error = -1;
		ntl_error_str = "Entry not found.";
		return ntl_error;
	}

	attr = ntl_entry(ld, entry);
	(void)ldap_msgfree(res);

	if (attr == NULL) {
		ntl_error = -2;
		ntl_error_str = "Entry empty?";
		return ntl_error;
	}

	if ((dnparts = ldap_explode_dn(source_dn, 0)) == NULL) {
		ntl_entry_free(attr);
		ntl_error = -10;
		ntl_error_str = "Source DN is bogus.";
		return ntl_error;
	} else if (dnparts[0] == NULL) {
		ntl_entry_free(attr);
		ldap_value_free(dnparts);
		ntl_error = -11;
		ntl_error_str = "Source DN is bogus.";
	}

	nt_asprintf(&dest_dn, "%s,%s", dnparts[0], dest_base);

	e = ntl_add(ld, dest_dn, &attr[1]); /* ignore first entry == DN */

	ntl_entry_free(attr);
	ldap_value_free(dnparts);
	free(dest_dn);

	return e;
}

int
ntl_copy(LDAP *ld, const char *source_dn, const char *dest_dn)
{
	/*
	 * Read source DN, convert entry to NTL_ATTR*,
	 * change DN, and create a new entry.
	 */

	LDAPMessage *res, *entry;
	NTL_ATTR *attr;
	char **dnparts, **odnparts;
	char *rdn_attr, *rdn_val, *ordn_attr, *ordn_val;
	int e;

	/*
	 * XXX/GCC: GCC4 Bug with -march=nocona
	 * warning: 'ordn_attr' may be used uninitialized in this function
	 * warning: 'rdn_attr' may be used uninitialized in this function
	 */
	rdn_attr = ordn_attr = NULL;

	if ((res = ntl_search(ld, source_dn, LDAP_SCOPE_BASE, NULL, NULL)) == NULL) {
		ntl_error = -1;
		ntl_error_str = "Entry not found.";
		return ntl_error;
	} else if ((entry = ldap_first_entry(ld, res)) == NULL) {
		(void)ldap_msgfree(res);
		ntl_error = -1;
		ntl_error_str = "Entry not found.";
		return ntl_error;
	}

	attr = ntl_entry(ld, entry);
	(void)ldap_msgfree(res);

	if (attr == NULL) {
		ntl_error = -2;
		ntl_error_str = "Entry empty?";
		return ntl_error;
	}

	/*
	 * Extract first component of new DN, add needed attribute.
	 *
	 * XXX: Only one defining attribute in first part allowed ATM!
	 *
	 * Try to be smart here: overwrite old attribute value with new if same attribute type.
	 */
	if ((dnparts = ldap_explode_dn(dest_dn, 0)) == NULL) {
		ntl_entry_free(attr);
		ntl_error = -4;
		ntl_error_str = "Destination DN is bogus.";
		return ntl_error;
	} else if ((odnparts = ldap_explode_dn(source_dn, 0)) == NULL) {
		ntl_entry_free(attr);
		ldap_value_free(dnparts);
		ntl_error = -5;
		ntl_error_str = "Source DN is bogus.";
		return ntl_error;
	}

	if (dnparts[0] == NULL || odnparts[0] == NULL) {
		e = ntl_error = -6;
		ntl_error_str = "Source and/or destination DN are bogus.";
		goto ntl_copy_out;
	}

	rdn_attr = nt_strdup(dnparts[0]);
	ordn_attr = nt_strdup(odnparts[0]);
	rdn_attr = strsep(&rdn_attr, "=");
	ordn_attr = strsep(&ordn_attr, "=");
	rdn_val = strchr(dnparts[0], '=') == NULL ? dnparts[0] : strchr(dnparts[0], '=') + 1;
	ordn_val = strchr(odnparts[0], '=') == NULL ? odnparts[0] : strchr(odnparts[0], '=') + 1;

	if (strcmp(rdn_attr, ordn_attr) == 0) {
		/* overwrite value for attribute */
		char **p;
		int i;

		i = 0;
		p = (char **)ntl_lookup(attr, rdn_attr);
		while (p[i] != NULL && strcmp(p[i], ordn_val) != 0)
			i++;

		if (p[i] == NULL) {
			e = ntl_error = -3;
			ntl_error_str = "Entry seems to have changed.";
			goto ntl_copy_out;
		}

		free(p[i]);
		p[i] = nt_strdup(rdn_val);
	} else
		ntl_attr_add(attr, rdn_attr, rdn_val);

	e = ntl_add(ld, dest_dn, &attr[1]); /* ignore first entry == DN */

ntl_copy_out:
	ntl_entry_free(attr);
	ldap_value_free(dnparts);
	ldap_value_free(odnparts);
	free(rdn_attr);
	free(ordn_attr);

	return e;
}

int
ntl_del(LDAP *ld, const char *dn)
{
	int r;

	r = ldap_delete_s(ld, dn);
	if (r == LDAP_SUCCESS)
		return 0;
	return r;
}

int
ntl_move_base(LDAP *ld, const char *source_dn, const char *dest_base)
{
	char **dnparts;
	char *dest_dn;
	int e;

	if ((dnparts = ldap_explode_dn(source_dn, 0)) == NULL) {
		ntl_error = -10;
		ntl_error_str = "Source DN is bogus.";
		return ntl_error;
	} else if (dnparts[0] == NULL) {
		ldap_value_free(dnparts);
		ntl_error = -11;
		ntl_error_str = "Source DN is bogus.";
		return ntl_error;
	}

	nt_asprintf(&dest_dn, "%s,%s", dnparts[0], dest_base);

	if ((e = ntl_copy(ld, source_dn, dest_dn)) == LDAP_SUCCESS)
		e = ntl_del(ld, source_dn);

	ldap_value_free(dnparts);
	free(dest_dn);

	return e;
}

int
ntl_move(LDAP *ld, const char *source_dn, const char *dest_dn)
{
	int e;

	if ((e = ntl_copy(ld, source_dn, dest_dn)) == LDAP_SUCCESS)
		e = ntl_del(ld, source_dn);
	return e;
}

char **
ntl_ldap_err(const char *msg, const char *dn, LDAP *ld, int e)
{
	char **out;
	char *info;

	out = nt_calloc(10, sizeof(char *));

	out[0] = nt_strdup(msg);
	out[1] = nt_strdup("DN:");
	out[2] = nt_strdup(dn);
	out[3] = nt_strdup("LDAP Error Message:");
	out[4] = nt_strdup(ldap_err2string(e));

	out[5] = nt_strdup("Matched DN:");
	info = NULL;
	(void)ldap_get_option(ld, LDAP_OPT_MATCHED_DN, &info);
	if (info != NULL) {
		out[6] = nt_strdup(info);
		ldap_memfree(info);
	} else
		out[6] = nt_strdup("");

	out[7] = nt_strdup("Additional Info:");
	info = NULL;
	(void)ldap_get_option(ld, LDAP_OPT_ERROR_STRING, &info);
	if (info != NULL) {
		out[8] = nt_strdup(info);
		ldap_memfree(info);
	} else
		out[8] = nt_strdup("");

	out[9] = NULL;

	return out;
}

void
ntl_ldap_err_free(char **s)
{
	int i;

	for (i = 0; s[i] != NULL; i++)
		free(s[i]);
	free(s);
}

static void
ntl_attr_add_generic(NTL_ATTR *set, const char *name, const char *value, int nocheck)
{
	int i, j;

	/*
	 * Find key and add value to key, if key exists.
	 */
	for (i = 0; i < set[0].tsize - 1 && set[i].name != NULL; i++) {
		if (strcasecmp(set[i].name, name) == 0) {
			/* go to end */
			j = 0;
			while (j < set[i].vsize - 1 && set[i].val[j] != NULL) {
				if (!nocheck && strcmp(set[i].val[j], value) == 0) {
					/* already there, ignore */
					return;
				}
				j++;
			}
			if (set[i].val[j] != NULL)
				goto ntl_attr_add_generic_fail;

			set[i].val[j] = nt_strdup(value);
			set[i].val[j+1] = NULL;
			return;
		}
	}
	if (set[i].name == NULL) {
		/* generate new key */
		set[i].name = nt_strdup(name);
		set[i].val = nt_calloc(NTL_MAX_VALS, sizeof(char *));
		set[i].vsize = NTL_MAX_VALS;
		set[i].val[0] = nt_strdup(value);
		set[i].val[1] = NULL;
		set[i+1].name = NULL;
		set[i+1].val = NULL;
		return;
	}

ntl_attr_add_generic_fail:
	errx(42, "NTL_ATTR full");
}

void
ntl_attr_add(NTL_ATTR *set, const char *name, const char *value)
{
	ntl_attr_add_generic(set, name, value, 0);
}

void
ntl_attr_add_nocheck(NTL_ATTR *set, const char *name, const char *value)
{
	ntl_attr_add_generic(set, name, value, 1);
}

char *
ntl_strip(const char *s)
{
	char buf[65536];
	int i;

	if (strlen(s) >= sizeof(buf) - 1)
		return nt_strdup(s);

	/* ignore leading whitespace */
	while (*s != '\0' && *s == ' ')
		s++;
	(void)strncpy(buf, s, sizeof(buf) - 1);
	buf[sizeof(buf) - 1] = '\0';

	/* remove trailing whitespace */
	i = strlen(buf);
	while (i > 1 && buf[i-1] == ' ')
		i--;
	buf[i] = '\0';

	return nt_strdup(buf);
}

void
ntl_in_nl(char *s)
{
	int i;

	if (s == NULL)
		return;

	for (i = 0; s[i] != '\0'; i++)
		if (s[i] == '')
			s[i] = '\n';
}

void
ntl_out_nl(char *s)
{
	int i;

	if (s == NULL)
		return;

	for (i = 0; s[i] != '\0'; i++)
		if (s[i] == '\n')
			s[i] = '';
}
