/*
 * $Id: 30d436e6ad237838a22db4c089aa9e41477d0d3f $
 *
 * This file is part of ntldap.
 *
 * Copyright (c) 2006, 2007, 2008 Niklas Olmes <niklas@noxa.de>
 *                                http://www.nevm.de/ntldap/
 *                                NEVM-RIPE
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "ldap.h"

#include "ntl.h"
#include "ntlc.h"
#include "nt_alloc.h"
#include "progname.h"

extern int ntl_error;

static const char rcsid[] = "$Id: 30d436e6ad237838a22db4c089aa9e41477d0d3f $";

static void usage(void);

static void
usage(void)
{
	(void)printf("Usage: %s [-h] [ -f cfg_file ] [ -p profile ] [ -b basedn ] [ -s base|subtree|onelevel ] filter [attributes]\n\n", getprogname());
	(void)printf("Examples:\n     %s 'objectClass=*'\n     ntlsearch -b 'ou=noxa.de,o=mail,dc=noxa,dc=de' 'uid=hostmaster'\n", getprogname());
	(void)printf("     %s -s base -p noxa 'objectClass=*' 'subschemaSubentry' 'objectClass' 'dc' 'modifyTimestamp'\n", getprogname());
	(void)printf("     %s -s base -p noxa -b 'cn=Subschema' 'objectClass=*' '+'\n\n", getprogname());
	(void)printf("Return Codes:\n     %d => success\n     1 => connection error\n     2 => search error\n     3 => no entries\n     126 => config file\n     127 => usage\n\n", EXIT_SUCCESS);

	(void)printf("Copyright (c) 2006, 2007, 2008 Niklas Olmes <niklas@noxa.de>\n"
		     "                               http://www.nevm.de/ntldap/\n"
		     "                               NEVM-RIPE\n"
		     "Licensed under the GNU GPL Version 2.\n");
}

int
main(int argc, char *argv[])
{
	LDAP *ld;
	LDAPMessage *res, *entry;
	NTL_CFG *cfg;
	char profile[256], cfgfile[256];
	char *filter, *basedn;
	char **attributes;
	char **e_msg;
	int scope;
	int e, c, i;

	setprogname(argv[0]);

	if (argc <= 1) {
		usage();
		return 127;
	}

	ld = NULL;
	res = entry = NULL;
	cfg = NULL;
	*profile = *cfgfile = '\0';
	filter = basedn = NULL;
	attributes = NULL;
	e_msg = NULL;
	scope = LDAP_SCOPE_SUBTREE;

	while ((c = getopt(argc, argv, "f:p:b:s:h")) != -1) {
		switch(c) {
		case 'f':
			(void)snprintf(cfgfile, sizeof(cfgfile), "%s", optarg);
			break;
		case 'p':
			(void)snprintf(profile, sizeof(profile), "%s", optarg);
			break;
		case 'b':
			basedn = nt_strdup(optarg);
			break;
		case 'h':
			usage();
			e = 127;
			goto cleanup;
		case 's':
			if (strcmp(optarg, "base") == 0)
				scope = LDAP_SCOPE_BASE;
			else if (strcmp(optarg, "subtree") == 0)
				scope = LDAP_SCOPE_SUBTREE;
			else if (strcmp(optarg, "onelevel") == 0)
				scope = LDAP_SCOPE_ONELEVEL;
			else {
				(void)fprintf(stderr, "Error: Unknwon scope \"%s\".\n", optarg);
				e = 127;
				goto cleanup;
			}
			break;
		default:
			e = 127;
			goto cleanup;
		}
	}
	argc -= optind;
	argv += optind;

	if (argc < 1) {
		(void)fprintf(stderr, "Error: Need filter.\n");
		e = 127;
		goto cleanup;
	}
	filter = argv[0];

	if (argc > 1) {
		argv++;
		attributes = nt_calloc((unsigned int)argc + 1, sizeof(char *));
		for (i = 0; i < argc - 1; i++)
			attributes[i] = nt_strdup(argv[i]);
		attributes[argc - 1] = NULL;
	}

	if ((cfg = ntlc_cfg_read(*cfgfile == '\0' ? NULL : cfgfile, *profile == '\0' ? NULL : profile)) == NULL) {
		ntlc_perror(argv[0]);
		e = 126;
		goto cleanup;
	}

	if (basedn == NULL)
		basedn = nt_strdup(cfg->basedn);

	e = EXIT_SUCCESS;
	if ((ld = ntl_connect(cfg)) == NULL) {
		e = 1;
		e_msg = ntl_ldap_err("Connection failed:", basedn, ld, ntl_error);
	} else if ((res = ntl_search(ld, basedn, scope, filter, (const char **)attributes)) == NULL) {
		e = 2;
		e_msg = ntl_ldap_err("Search error:", basedn, ld, ntl_error);
	} else if ((entry = ldap_first_entry(ld, res)) == NULL) {
		e = 3;
		e_msg = ntl_ldap_err("Nothing found:", basedn, ld, ntl_error);
	} else {
		NTL_ATTR *attr;
		char **ldif;
		char *p;
		int j;

		for (
			i = 1;
			entry != NULL;
			entry = ldap_next_entry(ld, entry), i++
		) {
			attr = ntl_entry(ld, entry);
			ldif = ntl_ldif_entry(attr);
			for (j = 0; ldif[j] != NULL; j++) {
				p = ntl_strip(ldif[j]);
				(void)printf("%s\n", p);
				free(p);
			}
			ntl_entry_free(attr);
			ntl_ldif_free(ldif);
			(void)printf("\n\n");
		}
	}

	if (e_msg != NULL) {
		(void)fprintf(stderr, "%s\n", e_msg[0]);
		(void)fprintf(stderr, ".--------------------------------------------------\n");
		for (i = 1; e_msg[i] != NULL && e_msg[i+1] != NULL; i += 2)
			(void)fprintf(stderr, "| %s %s\n", e_msg[i], e_msg[i+1]);
		(void)fprintf(stderr, "`--------------------------------------------------\n");
		ntl_ldap_err_free(e_msg);
	}

cleanup:
	if (basedn != NULL)
		free(basedn);
	if (attributes != NULL) {
		for (i = 0; attributes[i] != NULL; i++)
			free(attributes[i]);
		free(attributes);
	}

	if (cfg != NULL)
		ntlc_cfg_free(cfg);

	if (res != NULL)
		(void)ldap_msgfree(res);
	if (ld != NULL)
		(void)ldap_unbind(ld);

	return e;
}
